# Copyright (c) Pymatgen Development Team.
# Distributed under the terms of the MIT License.

import os
import shutil
import time
import unittest
import warnings
from glob import glob

from pymatgen.io.cp2k.inputs import Keyword, KeywordList
from pymatgen.io.cp2k.sets import StaticSet

from custodian.cp2k.handlers import (
    AbortHandler,
    FrozenJobErrorHandler,
    NumericalPrecisionHandler,
    StdErrHandler,
    UnconvergedScfErrorHandler,
    get_conv,
)
from custodian.cp2k.interpreter import Cp2kModder
from tests.conftest import TEST_FILES

TEST_FILES_DIR = f"{TEST_FILES}/cp2k"


def clean_dir(dct) -> None:
    for file in glob(os.path.join(dct, "error.*.tar.gz")):
        os.remove(file)
    for file in glob(os.path.join(dct, "custodian.chk.*.tar.gz")):
        os.remove(file)


class HandlerTests(unittest.TestCase):
    def setUp(self) -> None:
        warnings.filterwarnings("ignore")

        clean_dir(TEST_FILES_DIR)

        time.sleep(1)  # for frozenhandler

        shutil.copy(f"{TEST_FILES_DIR}/cp2k.inp.orig", f"{TEST_FILES_DIR}/cp2k.inp")

        self.input_file = f"{TEST_FILES_DIR}/cp2k.inp"

        self.output_file_preconditioner = f"{TEST_FILES_DIR}/cp2k.out.precondstuck"
        self.output_file_cholesky = f"{TEST_FILES_DIR}/cp2k.out.cholesky"
        self.output_file_imprecise = f"{TEST_FILES_DIR}/cp2k.out.imprecise"
        self.output_file_unconverged = f"{TEST_FILES_DIR}/cp2k.out.unconverged"
        self.output_file_stderr = f"{TEST_FILES_DIR}/std_err.txt"
        self.output_file_hybrid = f"{TEST_FILES_DIR}/cp2k.out.hybrid"
        self.output_file_conv = f"{TEST_FILES_DIR}/cp2k.out.conv"

        self.modder = Cp2kModder(filename=self.input_file)

    def test(self) -> None:
        """Ensure modder works"""
        kwds = KeywordList(keywords=[Keyword("BASIS_SET_FILE_NAME", "FILE1"), Keyword("BASIS_SET_FILE_NAME", "FILE2")])
        actions = [
            {"dict": self.input_file, "action": {"_set": {"FORCE_EVAL": {"METHOD": "NOT QA"}}}},
            {"dict": self.input_file, "action": {"_set": {"FORCE_EVAL": {"DFT": {"BASIS_SET_FILE_NAME": kwds}}}}},
            {
                "dict": self.input_file,
                "action": {"_set": {"FORCE_EVAL": {"DFT": {"SCF": {"MAX_SCF": 50}, "OUTER_SCF": {"MAX_SCF": 8}}}}},
            },
        ]
        self.modder.apply_actions(actions=actions)
        assert self.modder.ci["FORCE_EVAL"]["METHOD"] == Keyword("METHOD", "NOT QA")
        assert isinstance(self.modder.ci["FORCE_EVAL"]["DFT"]["BASIS_SET_FILE_NAME"], KeywordList)

    def test_handler_inits(self) -> None:
        """Ensure handlers initialize fine without real input/output files"""
        for handler in (AbortHandler, FrozenJobErrorHandler, NumericalPrecisionHandler, UnconvergedScfErrorHandler):
            handler()

    def test_frozenjobhandler(self) -> None:
        """Handler for frozen job"""
        handler = FrozenJobErrorHandler(
            input_file=self.input_file, output_file=self.output_file_preconditioner, timeout=1
        )
        assert handler.check()
        ci = StaticSet.from_file(self.input_file)
        assert ci["FORCE_EVAL"]["DFT"]["SCF"]["OT"]["PRECONDITIONER"] == Keyword(
            "PRECONDITIONER", "FULL_SINGLE_INVERSE"
        )
        handler.correct()

        ci = StaticSet.from_file(self.input_file)
        assert ci["FORCE_EVAL"]["DFT"]["SCF"]["OT"]["PRECONDITIONER"] == Keyword("PRECONDITIONER", "FULL_ALL")

        handler = FrozenJobErrorHandler(
            input_file=self.input_file, output_file=self.output_file_preconditioner, timeout=1
        )
        assert handler.check()
        handler.correct()
        ci = StaticSet.from_file(self.input_file)
        assert ci["FORCE_EVAL"]["DFT"]["SCF"]["OT"]["PRECOND_SOLVER"] == Keyword("PRECOND_SOLVER", "DIRECT")

        handler = FrozenJobErrorHandler(input_file=self.input_file, output_file=self.output_file_imprecise, timeout=1)
        handler.check()

    def test_unconverged_handler(self) -> None:
        """Handler for SCF handling not working"""
        ci = StaticSet.from_file(self.input_file)
        handler = UnconvergedScfErrorHandler(input_file=self.input_file, output_file=self.output_file_unconverged)
        handler.check()
        assert handler.is_ot
        assert ci["force_eval"]["dft"]["scf"]["ot"]["minimizer"] == Keyword("MINIMIZER", "DIIS")
        actions = handler.correct()
        assert actions["errors"], ["Non-converging Job"]
        ci = StaticSet.from_file(self.input_file)
        assert ci["force_eval"]["dft"]["scf"]["ot"]["minimizer"] == Keyword("MINIMIZER", "CG")

        # Fake diag check. Turns on mixing
        handler.is_ot = False
        actions = handler.correct()
        assert actions["errors"], ["Non-converging Job"]
        ci = StaticSet.from_file(self.input_file)
        assert ci["force_eval"]["dft"]["scf"]["MIXING"]["ALPHA"] == Keyword("ALPHA", 0.1)

    def test_abort_handler(self) -> None:
        """Checks if cp2k called abort"""
        handler = AbortHandler(input_file=self.input_file, output_file=self.output_file_cholesky)
        assert handler.check()

    def test_imprecision_handler(self) -> None:
        """Check for low precision leading to stagnant SCF"""
        handler = NumericalPrecisionHandler(self.input_file, output_file=self.output_file_imprecise, max_same=3)
        assert handler.check()
        c = handler.correct()
        assert c["errors"], ["Insufficient precision"]

    def test_std_out(self) -> None:
        """Errors sent to the std out instead of cp2k out"""
        handler = StdErrHandler(std_err=self.output_file_stderr)
        assert handler.check()
        handler.correct()

    def test_conv(self) -> None:
        """Check that SCF convergence can be read"""
        assert len(get_conv(self.output_file_conv)) == 45



"""Created 17 June, 2024"""

import os

import numpy as np
import pytest
from monty.os.path import zpath
from pymatgen.core import Lattice, Structure
from pymatgen.io.vasp import Kpoints
from pymatgen.util.testing import MatSciTest

from custodian.vasp.utils import _estimate_num_k_points_from_kspacing, increase_k_point_density
from tests.conftest import TEST_FILES


class TestKPointUtils(MatSciTest):
    large_structure = Structure.from_file(zpath(os.path.join(TEST_FILES, "POSCAR_mp-1200292")))

    small_structure = Structure(
        Lattice.from_parameters(a=3.8, b=3.8, c=3.8, alpha=60.0, beta=60.0, gamma=60.0),
        ["Si", "Si"],
        [[(-1) ** i * 0.125 for _ in range(3)] for i in range(2)],
    )

    def test_k_point_estimate(self) -> None:
        kspacing_values = [1, 0.8, 0.6, 0.4, 0.2, 0.1]
        expected_kpoints_per_axis = [(1, 1, 1), (1, 1, 1), (1, 1, 1), (2, 1, 1), (3, 2, 2), (5, 4, 4)]
        assert all(
            _estimate_num_k_points_from_kspacing(self.large_structure, kspacing) == expected_kpoints_per_axis[i]
            for i, kspacing in enumerate(kspacing_values)
        )

    def test_kpoint_density_increase(self):
        new_kpoints = increase_k_point_density(Kpoints(), self.large_structure, force_gamma=True, min_kpoints=4)
        assert new_kpoints["kpoints"] == ((2, 2, 1),)

        new_kpoints = increase_k_point_density(2, self.large_structure, force_gamma=True, min_kpoints=10)
        assert new_kpoints["KSPACING"] == pytest.approx(0.229885)
        new_kpoints_divis = _estimate_num_k_points_from_kspacing(self.large_structure, new_kpoints["KSPACING"])
        assert np.prod(new_kpoints_divis) == 12

        new_kpoints = increase_k_point_density(Kpoints(), self.small_structure, force_gamma=True, min_kpoints=14)
        assert new_kpoints["kpoints"] == ((3, 3, 3),)



"""This module implements basic kinds of jobs for FEFF runs."""

import logging
import os
import shutil
import subprocess

from monty.shutil import decompress_dir

from custodian.custodian import Job
from custodian.utils import backup

logger = logging.getLogger(__name__)

__author__ = "Chen Zheng"
__version__ = "0.1"
__maintainer__ = "Chen Zheng"
__email__ = "chz022@ucsd.edu"
__status__ = "Alpha"
__date__ = "10/20/17"

FEFF_INPUT_FILES = {"feff.inp"}
FEFF_BACKUP_FILES = {"ATOMS", "HEADER", "PARAMETERS", "POTENTIALS"}


class FeffJob(Job):
    """A basic FEFF job, run whatever is in the directory."""

    def __init__(
        self,
        feff_cmd,
        output_file="feff.out",
        stderr_file="std_feff_err.txt",
        backup=True,
        gzipped=False,
        gzipped_prefix="feff_out",
    ) -> None:
        """
        This constructor is used for a standard FEFF initialization.

        Args:
            feff_cmd (str): the name of the full executable for running FEFF
            output_file (str): Name of file to direct standard out to.
                Defaults to "feff.out".
            stderr_file (str): Name of file direct standard error to.
                Defaults to "std_feff_err.txt".
            backup (bool): Indicating whether to backup the initial input files.
                If True, the feff.inp will be copied with a ".orig" appended.
                Defaults to True.
            gzipped (bool): Whether to gzip the final output. Defaults to False.
            gzipped_prefix (str): prefix to the feff output files archive. Defaults
                to feff_out, which means a series of feff_out.1.tar.gz, feff_out.2.tar.gz, ...
                will be generated.
        """
        self.feff_cmd = feff_cmd
        self.output_file = output_file
        self.stderr_file = stderr_file
        self.backup = backup
        self.gzipped = gzipped
        self.gzipped_prefix = gzipped_prefix

    def setup(self, directory="./") -> None:
        """Performs initial setup for FeffJob, do backing up."""
        decompress_dir(directory)

        if self.backup:
            for file in FEFF_INPUT_FILES:
                shutil.copy(os.path.join(directory, file), os.path.join(directory, f"{file}.orig"))

            for file in FEFF_BACKUP_FILES:
                if os.path.isfile(os.path.join(directory, file)):
                    shutil.copy(os.path.join(directory, file), os.path.join(directory, f"{file}.orig"))

    def run(self, directory="./"):
        """
        Performs the actual FEFF run
        Returns:
            (subprocess.Popen) Used for monitoring.
        """
        with (
            open(os.path.join(directory, self.output_file), "w") as f_std,
            open(os.path.join(directory, self.stderr_file), "w", buffering=1) as f_err,
        ):
            # Use line buffering for stderr
            # On TSCC, need to run shell command
            return subprocess.Popen(self.feff_cmd, cwd=directory, stdout=f_std, stderr=f_err, shell=True)  # pylint: disable=R1732

    def postprocess(self, directory="./") -> None:
        """Renaming or gzipping all the output as needed."""
        if self.gzipped:
            backup("*", directory=directory, prefix=self.gzipped_prefix)



"""
pymatgen documentation build configuration file, created by
sphinx-quickstart on Tue Nov 15 00:13:52 2011.

This file is execfile()d with the current directory set to its containing dir.

Note that not all possible configuration values are present in this
autogenerated file.

All configuration values have a default; values that are commented out
serve to show the default.
"""

from __future__ import annotations

from pymatgen.core import __author__

project = "pymatgen"
copyright = "2022, Materials Virtual Lab"
author = __author__


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
# Napoleon is necessary to parse Google style docstrings. Markdown builder allows the generation of markdown output.
extensions = ["sphinx.ext.autodoc", "sphinx.ext.napoleon", "myst_parser", "sphinx_markdown_builder"]
# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]
autoclass_content = "both"

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = []


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "alabaster"

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]

myst_heading_anchors = 3

autodoc_default_options = {"private-members": False}



import os

from monty.os import cd
from monty.tempfile import ScratchDir
from pymatgen.io.feff.inputs import Atoms, Tags

from custodian.feff.jobs import FeffJob
from tests.conftest import TEST_FILES

__author__ = "Chen Zheng"
__copyright__ = "Copyright 2012, The Materials Project"
__version__ = "0.1"
__maintainer__ = "Chen Zheng"
__email__ = "chz022@ucsd.edu"
__date__ = "Oct 18, 2017"

TEST_DIR = f"{TEST_FILES}/feff_unconverged"


def test_as_from_dict() -> None:
    f = FeffJob("hello")
    f2 = FeffJob.from_dict(f.as_dict())
    assert isinstance(f, FeffJob)
    assert f2.feff_cmd == "hello"


def test_setup() -> None:
    with cd(TEST_DIR), ScratchDir(".", copy_from_current_on_enter=True):
        f = FeffJob("hello", backup=True)
        f.setup()

        parameter = Tags.from_file("feff.inp")
        parameter_orig = Tags.from_file("feff.inp.orig")
        assert parameter == parameter_orig

        atom = Atoms.cluster_from_file("feff.inp")
        atom_origin = Atoms.cluster_from_file("feff.inp.orig")
        assert atom == atom_origin


def test_postprocess() -> None:
    with cd(TEST_DIR), ScratchDir(".", copy_from_current_on_enter=True):
        f = FeffJob("hello", backup=True, gzipped=True)
        f.postprocess()
        assert os.path.isfile("feff_out.1.tar.gz")
        f.postprocess()
        assert os.path.isfile("feff_out.2.tar.gz")



"""This module mocks functions needed for pytest."""

import multiprocessing
import os

import pytest

TEST_DIR = os.path.dirname(__file__)
TEST_FILES = f"{TEST_DIR}/files"


@pytest.fixture(autouse=True)
def _patch_get_potential_energy(monkeypatch) -> None:
    """Monkeypatch the multiprocessing.cpu_count() function to always return 64."""
    monkeypatch.setattr(multiprocessing, "cpu_count", lambda: 64)



"""
Custodian is a simple, robust and flexible just-in-time (JIT) job management
framework written in Python.
"""

import os
from importlib.metadata import PackageNotFoundError, version

from .custodian import Custodian

__author__ = (
    "Shyue Ping Ong, William Davidson Richards, Stephen Dacek, Xiaohui Qu, Matthew Horton, "
    "Samuel M. Blau, Janosh Riebesell"
)
try:
    __version__ = version("custodian")
except PackageNotFoundError:  # pragma: no cover
    # package is not installed
    pass


PKG_DIR = os.path.dirname(__file__)
ROOT = os.path.dirname(PKG_DIR)



"""This module implements basic kinds of jobs for JDFTx runs."""

import logging
import os
import shlex
import subprocess

import psutil

from custodian.custodian import Job

logger = logging.getLogger(__name__)


class JDFTxJob(Job):
    """A basic JDFTx job. Runs whatever is in the working directory."""

    def __init__(
        self,
        jdftx_cmd,
        input_file="init.in",
        output_file="jdftx.out",
        stderr_file="std_err.txt",
    ) -> None:
        """
        Args:
            jdftx_cmd (str): Command to run JDFTx as a string.
            input_file (str): Name of the file to use as input to JDFTx
                executable. Defaults to "init.in"
            output_file (str): Name of file to direct standard out to.
                Defaults to "jdftx.out".
            stderr_file (str): Name of file to direct standard error to.
                Defaults to "std_err.txt".
        """
        self.jdftx_cmd = jdftx_cmd
        self.input_file = input_file
        self.output_file = output_file
        self.stderr_file = stderr_file

    def setup(self, directory="./") -> None:
        """No setup required."""

    def run(self, directory="./"):
        """
        Perform the actual JDFTx run.

        Returns:
        -------
            (subprocess.Popen) Used for monitoring.
        """
        cmd = self.jdftx_cmd + " -i " + self.input_file + " -o " + self.output_file
        logger.info(f"Running {cmd}")
        with (
            open(os.path.join(directory, self.output_file), "w") as f_std,
            open(os.path.join(directory, self.stderr_file), "w", buffering=1) as f_err,
        ):
            # use line buffering for stderr
            return subprocess.run(
                shlex.split(cmd),
                cwd=directory,
                stdout=f_std,
                stderr=f_err,
                shell=False,
                check=False,
            )

    def postprocess(self, directory="./") -> None:
        """No post-processing required."""

    def terminate(self, directory="./") -> None:
        """Terminate JDFTx."""
        work_dir = directory
        logger.info(f"Killing JDFTx processes in {work_dir=}.")
        for proc in psutil.process_iter():
            try:
                if "jdftx" in proc.name():
                    print("name:", proc.name())
                    open_paths = [file.path for file in proc.open_files()]
                    run_path = os.path.join(work_dir, self.output_file)
                    if (run_path in open_paths) and psutil.pid_exists(proc.pid):
                        self.terminate_process(proc)
                        return
            except (psutil.NoSuchProcess, psutil.AccessDenied) as e:
                logger.warning(f"Exception {e} encountered while killing JDFTx.")
                continue

        logger.warning(
            f"Killing JDFTx processes in {work_dir=} failed with subprocess.Popen.terminate(). Resorting to 'killall'."
        )
        cmd = self.jdftx_cmd
        print("cmd:", cmd)
        if "jdftx" in cmd:
            subprocess.run(["killall", f"{cmd}"], check=False)

    @staticmethod
    def terminate_process(proc, timeout=5):
        """Terminate a process gracefully, then forcefully if necessary."""
        try:
            proc.terminate()
            try:
                proc.wait(timeout=timeout)
            except psutil.TimeoutExpired:
                # If process is still running after the timeout, kill it
                logger.warning(f"Process {proc.pid} did not terminate gracefully, killing it.")
                proc.kill()
                # proc.wait()
            else:
                logger.info(f"Process {proc.pid} terminated gracefully.")
        except (psutil.NoSuchProcess, psutil.AccessDenied) as e:
            logger.warning(f"Error while terminating process {proc.pid}: {e}")



"""Tests for QChem Custodian"""

import os
import unittest

# skip QChem tests if openbabel is not installed but never skip them in CI
if "CI" not in os.environ:
    try:
        import openbabel
    except ImportError:
        raise unittest.SkipTest("Install openbabel to run the QChem tests in this directory.")
        # see .github/workflows/pytest.yml for install command

__author__ = "Samuel Blau, Brandon Wood, Shyam Dwaraknath"
__credits__ = "Xiaohui Qu"



import tarfile
from pathlib import Path

from custodian.utils import backup, tracked_lru_cache


def test_cache_and_clear() -> None:
    n_calls = 0

    @tracked_lru_cache
    def some_func(x):
        nonlocal n_calls
        n_calls += 1
        return x

    assert some_func(1) == 1
    assert n_calls == 1
    assert some_func(2) == 2
    assert n_calls == 2
    assert some_func(1) == 1
    assert n_calls == 2

    assert len(tracked_lru_cache.cached_functions) == 1

    tracked_lru_cache.tracked_cache_clear()

    assert len(tracked_lru_cache.cached_functions) == 0

    assert some_func(1) == 1
    assert n_calls == 3


def test_backup(tmp_path, monkeypatch) -> None:
    monkeypatch.chdir(tmp_path)
    with open("INCAR", "w") as f:
        f.write("This is a test file.")

    backup(["INCAR"])

    assert Path("error.1.tar.gz").exists()
    with tarfile.open("error.1.tar.gz", "r:gz") as tar:
        assert len(tar.getmembers()) == 1
        assert tar.getnames() == ["error.1/INCAR"]


def test_backup_with_glob(tmp_path, monkeypatch) -> None:
    monkeypatch.chdir(tmp_path)
    with open("INCAR", "w") as f:
        f.write("This is a test file.")
    with open("POSCAR", "w") as f:
        f.write("This is a test file.")
    with open("garbage", "w") as f:
        f.write("This is a test file.")

    backup(["*CAR"])

    assert Path("error.1.tar.gz").exists()
    with tarfile.open("error.1.tar.gz", "r:gz") as tar:
        assert len(tar.getmembers()) == 2
        assert set(tar.getnames()) == {"error.1/INCAR", "error.1/POSCAR"}


def test_backup_with_directory(tmp_path) -> None:
    with open(tmp_path / "INCAR", "w") as f:
        f.write("This is a test file.")

    backup(["INCAR"], directory=tmp_path)

    assert Path(tmp_path / "error.1.tar.gz").exists()
    with tarfile.open(tmp_path / "error.1.tar.gz", "r:gz") as tar:
        assert len(tar.getmembers()) == 1
        assert tar.getnames() == ["error.1/INCAR"]



import os
import shutil
import unittest
from unittest import TestCase
from unittest.mock import patch

import pytest
from pymatgen.io.qchem.inputs import QCInput

from custodian.qchem.jobs import QCJob
from tests.conftest import TEST_FILES

try:
    from openbabel import openbabel as ob
except ImportError:
    ob = None

__author__ = "Samuel Blau"
__copyright__ = "Copyright 2018, The Materials Project"
__version__ = "0.1"
__maintainer__ = "Samuel Blau"
__email__ = "samblau1@gmail.com"
__status__ = "Alpha"
__date__ = "6/6/18"
__credits__ = "Shyam Dwaraknath"

TEST_DIR = f"{TEST_FILES}/qchem/new_test_files"
SCR_DIR = f"{TEST_DIR}/scratch"
CWD = os.getcwd()
skip_if_no_openbabel = unittest.skipIf(ob is None, "openbabel not installed")


@skip_if_no_openbabel
class QCJobTest(TestCase):
    def setUp(self) -> None:
        os.makedirs(SCR_DIR)
        shutil.copyfile(f"{TEST_DIR}/no_nbo.qin", f"{SCR_DIR}/mol.qin")
        shutil.copyfile(f"{TEST_DIR}/nbo7.qin", f"{SCR_DIR}/different.qin")
        os.chdir(SCR_DIR)

    def tearDown(self) -> None:
        os.chdir(CWD)
        shutil.rmtree(SCR_DIR)

    def test_defaults(self) -> None:
        with patch("custodian.qchem.jobs.shutil.copy") as copy_patch:
            job = QCJob(qchem_command="qchem", max_cores=32)
            assert job.current_command == "qchem -nt 32 ./mol.qin ./mol.qout scratch"
            job.setup()
            assert copy_patch.call_args_list[0][0][0] == "./mol.qin"
            assert copy_patch.call_args_list[0][0][1] == "./mol.qin.orig"
            assert os.environ["QCSCRATCH"] == os.getcwd()
            assert os.environ["QCTHREADS"] == "32"
            assert os.environ["OMP_NUM_THREADS"] == "32"

    def test_not_defaults(self) -> None:
        job = QCJob(
            qchem_command="qchem -slurm",
            multimode="mpi",
            input_file="different.qin",
            output_file="not_default.qout",
            max_cores=12,
            calc_loc="/not/default/",
            nboexe="/path/to/nbo7.i4.exe",
            backup=False,
        )
        assert job.current_command == "qchem -slurm -np 12 ./different.qin ./not_default.qout scratch"
        job.setup()
        assert os.environ["QCSCRATCH"] == os.getcwd()
        assert os.environ["QCLOCALSCR"] == "/not/default/"
        assert os.environ["NBOEXE"] == "/path/to/nbo7.i4.exe"
        assert os.environ["KMP_INIT_AT_FORK"] == "FALSE"

    def test_save_scratch(self) -> None:
        with patch("custodian.qchem.jobs.shutil.copy") as copy_patch:
            job = QCJob(
                qchem_command="qchem -slurm",
                max_cores=32,
                calc_loc="/tmp/scratch",
                save_scratch=True,
            )
            assert job.current_command == "qchem -slurm -nt 32 ./mol.qin ./mol.qout scratch"
            job.setup()
            assert copy_patch.call_args_list[0][0][0] == "./mol.qin"
            assert copy_patch.call_args_list[0][0][1] == "./mol.qin.orig"
            assert os.environ["QCSCRATCH"] == os.getcwd()
            assert os.environ["QCTHREADS"] == "32"
            assert os.environ["OMP_NUM_THREADS"] == "32"
            assert os.environ["QCLOCALSCR"] == "/tmp/scratch"


@skip_if_no_openbabel
class OptFFComplexUnlinkedTest(TestCase):
    def setUp(self) -> None:
        os.makedirs(SCR_DIR)
        shutil.copyfile(f"{TEST_DIR}/FF_complex/mol.qin.opt_0", f"{SCR_DIR}/mol.qin")
        shutil.copyfile(f"{TEST_DIR}/FF_complex/mol.qout.opt_0", f"{SCR_DIR}/mol.qout.opt_0")
        shutil.copyfile(f"{TEST_DIR}/FF_complex/mol.qout.freq_0", f"{SCR_DIR}/mol.qout.freq_0")
        shutil.copyfile(f"{TEST_DIR}/FF_complex/mol.qin.freq_0", f"{SCR_DIR}/mol.qin.freq_0")
        os.chdir(SCR_DIR)

    def tearDown(self) -> None:
        os.chdir(CWD)
        shutil.rmtree(SCR_DIR)

    def test_OptFF(self) -> None:
        job = QCJob.opt_with_frequency_flattener(
            qchem_command="qchem",
            max_cores=32,
            input_file="mol.qin",
            output_file="mol.qout",
            max_molecule_perturb_scale=0.5,
            linked=False,
        )
        expected_next = QCJob(
            qchem_command="qchem",
            max_cores=32,
            multimode="openmp",
            input_file="mol.qin",
            output_file="mol.qout",
            suffix=".opt_0",
            backup=True,
        ).as_dict()
        assert next(job).as_dict() == expected_next
        expected_next = QCJob(
            qchem_command="qchem",
            max_cores=32,
            multimode="openmp",
            input_file="mol.qin",
            output_file="mol.qout",
            suffix=".freq_0",
            backup=False,
        ).as_dict()
        assert next(job).as_dict() == expected_next
        assert (
            QCInput.from_file(f"{TEST_DIR}/FF_complex/mol.qin.freq_0_ref").as_dict()
            == QCInput.from_file(os.path.join(SCR_DIR, "mol.qin")).as_dict()
        )
        expected_next = QCJob(
            qchem_command="qchem",
            max_cores=32,
            multimode="openmp",
            input_file="mol.qin",
            output_file="mol.qout",
            suffix=".opt_1",
            backup=False,
        ).as_dict()
        assert next(job).as_dict() == expected_next
        assert (
            QCInput.from_file(f"{TEST_DIR}/FF_complex/mol.qin.opt_1_unlinked").as_dict()
            == QCInput.from_file(os.path.join(SCR_DIR, "mol.qin")).as_dict()
        )


@skip_if_no_openbabel
class OptFFTestComplexLinkedChangeNsegTest(TestCase):
    def setUp(self) -> None:
        os.makedirs(SCR_DIR)
        shutil.copyfile(f"{TEST_DIR}/FF_complex/mol.qin.opt_0", f"{SCR_DIR}/mol.qin")
        shutil.copyfile(f"{TEST_DIR}/FF_complex/mol.qin.opt_0", f"{SCR_DIR}/mol.qin.opt_0")
        shutil.copyfile(f"{TEST_DIR}/FF_complex/mol.qin.freq_0", f"{SCR_DIR}/mol.qin.freq_0")
        shutil.copyfile(f"{TEST_DIR}/FF_complex/mol.qin.opt_1", f"{SCR_DIR}/mol.qin.opt_1")
        shutil.copyfile(f"{TEST_DIR}/FF_complex/mol.qin.freq_1", f"{SCR_DIR}/mol.qin.freq_1")
        shutil.copyfile(f"{TEST_DIR}/FF_complex/mol.qout.opt_0", f"{SCR_DIR}/mol.qout.opt_0")
        shutil.copyfile(f"{TEST_DIR}/FF_complex/mol.qout.freq_0", f"{SCR_DIR}/mol.qout.freq_0")
        shutil.copyfile(f"{TEST_DIR}/FF_complex/mol.qout.opt_1", f"{SCR_DIR}/mol.qout.opt_1")
        shutil.copyfile(f"{TEST_DIR}/FF_complex/mol.qout.freq_1", f"{SCR_DIR}/mol.qout.freq_1")
        os.chdir(SCR_DIR)

    def tearDown(self) -> None:
        os.chdir(CWD)
        shutil.rmtree(SCR_DIR)

    def test_OptFF(self) -> None:
        job = QCJob.opt_with_frequency_flattener(
            qchem_command="qchem",
            max_cores=32,
            input_file="mol.qin",
            output_file="mol.qout",
            linked=True,
        )
        expected_next = QCJob(
            qchem_command="qchem",
            max_cores=32,
            multimode="openmp",
            input_file="mol.qin",
            output_file="mol.qout",
            suffix=".opt_0",
            backup=True,
            save_scratch=True,
        ).as_dict()
        assert next(job).as_dict() == expected_next
        expected_next = QCJob(
            qchem_command="qchem",
            max_cores=32,
            multimode="openmp",
            input_file="mol.qin",
            output_file="mol.qout",
            suffix=".freq_0",
            backup=False,
            save_scratch=True,
        ).as_dict()
        assert next(job).as_dict() == expected_next
        assert (
            QCInput.from_file(f"{TEST_DIR}/FF_complex/mol.qin.freq_0_ref").as_dict()
            == QCInput.from_file(os.path.join(SCR_DIR, "mol.qin")).as_dict()
        )
        expected_next = QCJob(
            qchem_command="qchem",
            max_cores=32,
            multimode="openmp",
            input_file="mol.qin",
            output_file="mol.qout",
            suffix=".opt_1",
            backup=False,
            save_scratch=True,
        ).as_dict()
        assert next(job).as_dict() == expected_next
        assert (
            QCInput.from_file(f"{TEST_DIR}/FF_complex/mol.qin.opt_1").as_dict()
            == QCInput.from_file(os.path.join(SCR_DIR, "mol.qin")).as_dict()
        )
        expected_next = QCJob(
            qchem_command="qchem",
            max_cores=32,
            multimode="openmp",
            input_file="mol.qin",
            output_file="mol.qout",
            suffix=".freq_1",
            backup=False,
            save_scratch=True,
        ).as_dict()
        assert next(job).as_dict() == expected_next
        assert (
            QCInput.from_file(f"{TEST_DIR}/FF_complex/mol.qin.freq_1").as_dict()
            == QCInput.from_file(os.path.join(SCR_DIR, "mol.qin")).as_dict()
        )


@skip_if_no_openbabel
class OptFFTest(TestCase):
    def setUp(self) -> None:
        os.makedirs(SCR_DIR)
        shutil.copyfile(f"{TEST_DIR}/FF_working/test.qin", f"{SCR_DIR}/test.qin")
        shutil.copyfile(f"{TEST_DIR}/FF_working/test.qout.opt_0", f"{SCR_DIR}/test.qout.opt_0")
        shutil.copyfile(f"{TEST_DIR}/FF_working/test.qout.freq_0", f"{SCR_DIR}/test.qout.freq_0")
        shutil.copyfile(f"{TEST_DIR}/FF_working/test.qin.freq_0", f"{SCR_DIR}/test.qin.freq_0")
        shutil.copyfile(f"{TEST_DIR}/FF_working/test.qin.freq_1", f"{SCR_DIR}/test.qin.freq_1")
        shutil.copyfile(f"{TEST_DIR}/FF_working/test.qout.opt_1", f"{SCR_DIR}/test.qout.opt_1")
        shutil.copyfile(f"{TEST_DIR}/FF_working/test.qout.freq_1", f"{SCR_DIR}/test.qout.freq_1")
        os.chdir(SCR_DIR)

    def tearDown(self) -> None:
        os.chdir(CWD)
        shutil.rmtree(SCR_DIR)

    def test_OptFF(self) -> None:
        job = QCJob.opt_with_frequency_flattener(
            qchem_command="qchem",
            max_cores=32,
            input_file="test.qin",
            output_file="test.qout",
            linked=False,
        )
        expected_next = QCJob(
            qchem_command="qchem",
            max_cores=32,
            multimode="openmp",
            input_file="test.qin",
            output_file="test.qout",
            suffix=".opt_0",
            backup=True,
        ).as_dict()
        assert next(job).as_dict() == expected_next
        expected_next = QCJob(
            qchem_command="qchem",
            max_cores=32,
            multimode="openmp",
            input_file="test.qin",
            output_file="test.qout",
            suffix=".freq_0",
            backup=False,
        ).as_dict()
        assert next(job).as_dict() == expected_next
        assert (
            QCInput.from_file(f"{TEST_DIR}/FF_working/test.qin.freq_0").as_dict()
            == QCInput.from_file(os.path.join(SCR_DIR, "test.qin")).as_dict()
        )
        expected_next = QCJob(
            qchem_command="qchem",
            max_cores=32,
            multimode="openmp",
            input_file="test.qin",
            output_file="test.qout",
            suffix=".opt_1",
            backup=False,
        ).as_dict()
        assert next(job).as_dict() == expected_next
        assert (
            QCInput.from_file(f"{TEST_DIR}/FF_working/test.qin.opt_1").as_dict()
            == QCInput.from_file(os.path.join(SCR_DIR, "test.qin")).as_dict()
        )
        expected_next = QCJob(
            qchem_command="qchem",
            max_cores=32,
            multimode="openmp",
            input_file="test.qin",
            output_file="test.qout",
            suffix=".freq_1",
            backup=False,
        ).as_dict()
        assert next(job).as_dict() == expected_next
        assert (
            QCInput.from_file(f"{TEST_DIR}/FF_working/test.qin.freq_1").as_dict()
            == QCInput.from_file(os.path.join(SCR_DIR, "test.qin")).as_dict()
        )
        with pytest.raises(StopIteration):
            next(job)


@skip_if_no_openbabel
class OptFFTest1(TestCase):
    def setUp(self) -> None:
        os.makedirs(SCR_DIR)
        shutil.copyfile(f"{TEST_DIR}/2620_complete/mol.qin.orig", f"{SCR_DIR}/mol.qin")
        shutil.copyfile(f"{TEST_DIR}/2620_complete/mol.qout.opt_0", f"{SCR_DIR}/mol.qout.opt_0")
        os.chdir(SCR_DIR)

    def tearDown(self) -> None:
        os.chdir(CWD)
        shutil.rmtree(SCR_DIR)

    def test_OptFF(self) -> None:
        job = QCJob.opt_with_frequency_flattener(
            qchem_command="qchem -slurm", max_cores=32, input_file="mol.qin", output_file="mol.qout", linked=False
        )
        expected_next = QCJob(
            qchem_command="qchem -slurm",
            max_cores=32,
            multimode="openmp",
            input_file="mol.qin",
            output_file="mol.qout",
            suffix=".opt_0",
            backup=True,
        ).as_dict()
        assert next(job).as_dict() == expected_next
        with pytest.raises(StopIteration):
            next(job)


@skip_if_no_openbabel
class OptFFTest2(TestCase):
    def setUp(self) -> None:
        os.makedirs(SCR_DIR)
        shutil.copyfile(f"{TEST_DIR}/disconnected_but_converged/mol.qin.orig", f"{SCR_DIR}/mol.qin")
        shutil.copyfile(f"{TEST_DIR}/disconnected_but_converged/mol.qout.opt_0", f"{SCR_DIR}/mol.qout.opt_0")
        shutil.copyfile(f"{TEST_DIR}/disconnected_but_converged/mol.qout.freq_0", f"{SCR_DIR}/mol.qout.freq_0")
        shutil.copyfile(f"{TEST_DIR}/disconnected_but_converged/mol.qin.freq_0", f"{SCR_DIR}/mol.qin.freq_0")
        os.chdir(SCR_DIR)

    def tearDown(self) -> None:
        os.chdir(CWD)
        shutil.rmtree(SCR_DIR)

    def test_OptFF(self) -> None:
        job = QCJob.opt_with_frequency_flattener(
            qchem_command="qchem -slurm",
            max_cores=32,
            input_file="mol.qin",
            output_file="mol.qout",
            linked=False,
        )
        expected_next = QCJob(
            qchem_command="qchem -slurm",
            max_cores=32,
            multimode="openmp",
            input_file="mol.qin",
            output_file="mol.qout",
            suffix=".opt_0",
            backup=True,
        ).as_dict()
        assert next(job).as_dict() == expected_next
        expected_next = QCJob(
            qchem_command="qchem -slurm",
            max_cores=32,
            multimode="openmp",
            input_file="mol.qin",
            output_file="mol.qout",
            suffix=".freq_0",
            backup=False,
        ).as_dict()
        assert next(job).as_dict() == expected_next
        assert (
            QCInput.from_file(f"{TEST_DIR}/disconnected_but_converged/mol.qin.freq_0").as_dict()
            == QCInput.from_file(os.path.join(SCR_DIR, "mol.qin")).as_dict()
        )
        with pytest.raises(StopIteration):
            next(job)


@skip_if_no_openbabel
class OptFFTestSwitching(TestCase):
    def setUp(self) -> None:
        os.makedirs(SCR_DIR)
        shutil.copyfile(f"{TEST_DIR}/FF_switching/mol.qin.orig", f"{SCR_DIR}/mol.qin")
        shutil.copyfile(f"{TEST_DIR}/FF_switching/mol.qout.opt_0", f"{SCR_DIR}/mol.qout.opt_0")
        shutil.copyfile(f"{TEST_DIR}/FF_switching/mol.qout.freq_0", f"{SCR_DIR}/mol.qout.freq_0")
        shutil.copyfile(f"{TEST_DIR}/FF_switching/mol.qin.freq_0", f"{SCR_DIR}/mol.qin.freq_0")
        shutil.copyfile(f"{TEST_DIR}/FF_switching/mol.qin.freq_1", f"{SCR_DIR}/mol.qin.freq_1")
        shutil.copyfile(f"{TEST_DIR}/FF_switching/mol.qin.freq_2", f"{SCR_DIR}/mol.qin.freq_2")
        shutil.copyfile(f"{TEST_DIR}/FF_switching/mol.qout.opt_1", f"{SCR_DIR}/mol.qout.opt_1")
        shutil.copyfile(f"{TEST_DIR}/FF_switching/mol.qout.freq_1", f"{SCR_DIR}/mol.qout.freq_1")
        shutil.copyfile(f"{TEST_DIR}/FF_switching/mol.qout.opt_2", f"{SCR_DIR}/mol.qout.opt_2")
        shutil.copyfile(f"{TEST_DIR}/FF_switching/mol.qout.freq_2", f"{SCR_DIR}/mol.qout.freq_2")
        os.chdir(SCR_DIR)

    def tearDown(self) -> None:
        os.chdir(CWD)
        shutil.rmtree(SCR_DIR)

    def test_OptFF(self) -> None:
        job = QCJob.opt_with_frequency_flattener(
            qchem_command="qchem -slurm",
            max_cores=32,
            input_file="mol.qin",
            output_file="mol.qout",
            linked=False,
        )
        expected_next = QCJob(
            qchem_command="qchem -slurm",
            max_cores=32,
            multimode="openmp",
            input_file="mol.qin",
            output_file="mol.qout",
            suffix=".opt_0",
            backup=True,
        ).as_dict()
        assert next(job).as_dict() == expected_next
        expected_next = QCJob(
            qchem_command="qchem -slurm",
            max_cores=32,
            multimode="openmp",
            input_file="mol.qin",
            output_file="mol.qout",
            suffix=".freq_0",
            backup=False,
        ).as_dict()
        assert next(job).as_dict() == expected_next
        assert (
            QCInput.from_file(f"{TEST_DIR}/FF_switching/mol.qin.freq_0").as_dict()
            == QCInput.from_file(os.path.join(SCR_DIR, "mol.qin")).as_dict()
        )
        expected_next = QCJob(
            qchem_command="qchem -slurm",
            max_cores=32,
            multimode="openmp",
            input_file="mol.qin",
            output_file="mol.qout",
            suffix=".opt_1",
            backup=False,
        ).as_dict()
        assert next(job).as_dict() == expected_next
        assert (
            QCInput.from_file(f"{TEST_DIR}/FF_switching/mol.qin.opt_1").as_dict()
            == QCInput.from_file(os.path.join(SCR_DIR, "mol.qin")).as_dict()
        )
        expected_next = QCJob(
            qchem_command="qchem -slurm",
            max_cores=32,
            multimode="openmp",
            input_file="mol.qin",
            output_file="mol.qout",
            suffix=".freq_1",
            backup=False,
        ).as_dict()
        assert next(job).as_dict() == expected_next
        assert (
            QCInput.from_file(f"{TEST_DIR}/FF_switching/mol.qin.freq_1").as_dict()
            == QCInput.from_file(os.path.join(SCR_DIR, "mol.qin")).as_dict()
        )
        expected_next = QCJob(
            qchem_command="qchem -slurm",
            max_cores=32,
            multimode="openmp",
            input_file="mol.qin",
            output_file="mol.qout",
            suffix=".opt_2",
            backup=False,
        ).as_dict()
        assert next(job).as_dict() == expected_next
        assert (
            QCInput.from_file(f"{TEST_DIR}/FF_switching/mol.qin.opt_2").as_dict()
            == QCInput.from_file(os.path.join(SCR_DIR, "mol.qin")).as_dict()
        )
        expected_next = QCJob(
            qchem_command="qchem -slurm",
            max_cores=32,
            multimode="openmp",
            input_file="mol.qin",
            output_file="mol.qout",
            suffix=".freq_2",
            backup=False,
        ).as_dict()
        assert next(job).as_dict() == expected_next
        assert (
            QCInput.from_file(f"{TEST_DIR}/FF_switching/mol.qin.freq_2").as_dict()
            == QCInput.from_file(os.path.join(SCR_DIR, "mol.qin")).as_dict()
        )
        with pytest.raises(StopIteration):
            next(job)


@skip_if_no_openbabel
class OptFFTest6004(TestCase):
    def setUp(self) -> None:
        os.makedirs(SCR_DIR)
        shutil.copyfile(f"{TEST_DIR}/6004_frag12/mol.qin.orig", f"{SCR_DIR}/mol.qin")
        shutil.copyfile(f"{TEST_DIR}/6004_frag12/mol.qout.opt_0", f"{SCR_DIR}/mol.qout.opt_0")
        shutil.copyfile(f"{TEST_DIR}/6004_frag12/mol.qout.freq_0", f"{SCR_DIR}/mol.qout.freq_0")
        shutil.copyfile(f"{TEST_DIR}/6004_frag12/mol.qin.freq_0", f"{SCR_DIR}/mol.qin.freq_0")
        shutil.copyfile(f"{TEST_DIR}/6004_frag12/mol.qin.freq_1", f"{SCR_DIR}/mol.qin.freq_1")
        shutil.copyfile(f"{TEST_DIR}/6004_frag12/mol.qin.freq_2", f"{SCR_DIR}/mol.qin.freq_2")
        shutil.copyfile(f"{TEST_DIR}/6004_frag12/mol.qout.opt_1", f"{SCR_DIR}/mol.qout.opt_1")
        shutil.copyfile(f"{TEST_DIR}/6004_frag12/mol.qout.freq_1", f"{SCR_DIR}/mol.qout.freq_1")
        shutil.copyfile(f"{TEST_DIR}/6004_frag12/mol.qout.opt_2", f"{SCR_DIR}/mol.qout.opt_2")
        shutil.copyfile(f"{TEST_DIR}/6004_frag12/mol.qout.freq_2", f"{SCR_DIR}/mol.qout.freq_2")
        os.chdir(SCR_DIR)

    def tearDown(self) -> None:
        os.chdir(CWD)
        shutil.rmtree(SCR_DIR)

    def test_OptFF(self) -> None:
        job = QCJob.opt_with_frequency_flattener(
            qchem_command="qchem -slurm",
            max_cores=32,
            input_file="mol.qin",
            output_file="mol.qout",
            linked=False,
        )
        expected_next = QCJob(
            qchem_command="qchem -slurm",
            max_cores=32,
            multimode="openmp",
            input_file="mol.qin",
            output_file="mol.qout",
            suffix=".opt_0",
            backup=True,
        ).as_dict()
        assert next(job).as_dict() == expected_next
        expected_next = QCJob(
            qchem_command="qchem -slurm",
            max_cores=32,
            multimode="openmp",
            input_file="mol.qin",
            output_file="mol.qout",
            suffix=".freq_0",
            backup=False,
        ).as_dict()
        assert next(job).as_dict() == expected_next
        assert (
            QCInput.from_file(f"{TEST_DIR}/6004_frag12/mol.qin.freq_0").as_dict()
            == QCInput.from_file(os.path.join(SCR_DIR, "mol.qin")).as_dict()
        )
        expected_next = QCJob(
            qchem_command="qchem -slurm",
            max_cores=32,
            multimode="openmp",
            input_file="mol.qin",
            output_file="mol.qout",
            suffix=".opt_1",
            backup=False,
        ).as_dict()
        assert next(job).as_dict() == expected_next
        assert (
            QCInput.from_file(f"{TEST_DIR}/6004_frag12/mol.qin.opt_1").as_dict()
            == QCInput.from_file(os.path.join(SCR_DIR, "mol.qin")).as_dict()
        )
        expected_next = QCJob(
            qchem_command="qchem -slurm",
            max_cores=32,
            multimode="openmp",
            input_file="mol.qin",
            output_file="mol.qout",
            suffix=".freq_1",
            backup=False,
        ).as_dict()
        assert next(job).as_dict() == expected_next
        assert (
            QCInput.from_file(f"{TEST_DIR}/6004_frag12/mol.qin.freq_1").as_dict()
            == QCInput.from_file(os.path.join(SCR_DIR, "mol.qin")).as_dict()
        )
        expected_next = QCJob(
            qchem_command="qchem -slurm",
            max_cores=32,
            multimode="openmp",
            input_file="mol.qin",
            output_file="mol.qout",
            suffix=".opt_2",
            backup=False,
        ).as_dict()
        assert next(job).as_dict() == expected_next
        assert (
            QCInput.from_file(f"{TEST_DIR}/6004_frag12/mol.qin.opt_2").as_dict()
            == QCInput.from_file(os.path.join(SCR_DIR, "mol.qin")).as_dict()
        )
        expected_next = QCJob(
            qchem_command="qchem -slurm",
            max_cores=32,
            multimode="openmp",
            input_file="mol.qin",
            output_file="mol.qout",
            suffix=".freq_2",
            backup=False,
        ).as_dict()
        assert next(job).as_dict() == expected_next
        assert (
            QCInput.from_file(f"{TEST_DIR}/6004_frag12/mol.qin.freq_2").as_dict()
            == QCInput.from_file(os.path.join(SCR_DIR, "mol.qin")).as_dict()
        )


@skip_if_no_openbabel
class OptFFTest5952(TestCase):
    def setUp(self) -> None:
        os.makedirs(SCR_DIR)
        shutil.copyfile(f"{TEST_DIR}/5952_frag16/mol.qin.orig", f"{SCR_DIR}/mol.qin")
        shutil.copyfile(f"{TEST_DIR}/5952_frag16/mol.qout.opt_0", f"{SCR_DIR}/mol.qout.opt_0")
        shutil.copyfile(f"{TEST_DIR}/5952_frag16/mol.qout.freq_0", f"{SCR_DIR}/mol.qout.freq_0")
        shutil.copyfile(f"{TEST_DIR}/5952_frag16/mol.qin.freq_0", f"{SCR_DIR}/mol.qin.freq_0")
        os.chdir(SCR_DIR)

    def tearDown(self) -> None:
        os.chdir(CWD)
        shutil.rmtree(SCR_DIR)

    def test_OptFF(self) -> None:
        job = QCJob.opt_with_frequency_flattener(
            qchem_command="qchem -slurm",
            max_cores=32,
            input_file="mol.qin",
            output_file="mol.qout",
            linked=False,
        )
        expected_next = QCJob(
            qchem_command="qchem -slurm",
            max_cores=32,
            multimode="openmp",
            input_file="mol.qin",
            output_file="mol.qout",
            suffix=".opt_0",
            backup=True,
        ).as_dict()
        assert next(job).as_dict() == expected_next
        expected_next = QCJob(
            qchem_command="qchem -slurm",
            max_cores=32,
            multimode="openmp",
            input_file="mol.qin",
            output_file="mol.qout",
            suffix=".freq_0",
            backup=False,
        ).as_dict()
        assert next(job).as_dict() == expected_next
        assert (
            QCInput.from_file(f"{SCR_DIR}/mol.qin.freq_0").as_dict()
            == QCInput.from_file(os.path.join(SCR_DIR, "mol.qin")).as_dict()
        )
        with pytest.raises(StopIteration):
            next(job)


@skip_if_no_openbabel
class OptFFTest5690(TestCase):
    def setUp(self) -> None:
        os.makedirs(SCR_DIR)
        shutil.copyfile(f"{TEST_DIR}/5690_frag18/mol.qin.orig", f"{SCR_DIR}/mol.qin")
        shutil.copyfile(f"{TEST_DIR}/5690_frag18/mol.qout.opt_0", f"{SCR_DIR}/mol.qout.opt_0")
        shutil.copyfile(f"{TEST_DIR}/5690_frag18/mol.qout.freq_0", f"{SCR_DIR}/mol.qout.freq_0")
        shutil.copyfile(f"{TEST_DIR}/5690_frag18/mol.qin.freq_0", f"{SCR_DIR}/mol.qin.freq_0")
        shutil.copyfile(f"{TEST_DIR}/5690_frag18/mol.qin.freq_1", f"{SCR_DIR}/mol.qin.freq_1")
        shutil.copyfile(f"{TEST_DIR}/5690_frag18/mol.qin.freq_2", f"{SCR_DIR}/mol.qin.freq_2")
        shutil.copyfile(f"{TEST_DIR}/5690_frag18/mol.qout.opt_1", f"{SCR_DIR}/mol.qout.opt_1")
        shutil.copyfile(f"{TEST_DIR}/5690_frag18/mol.qout.freq_1", f"{SCR_DIR}/mol.qout.freq_1")
        shutil.copyfile(f"{TEST_DIR}/5690_frag18/mol.qout.opt_2", f"{SCR_DIR}/mol.qout.opt_2")
        shutil.copyfile(f"{TEST_DIR}/5690_frag18/mol.qout.freq_2", f"{SCR_DIR}/mol.qout.freq_2")
        os.chdir(SCR_DIR)

    def tearDown(self) -> None:
        os.chdir(CWD)
        shutil.rmtree(SCR_DIR)

    def test_OptFF(self) -> None:
        job = QCJob.opt_with_frequency_flattener(
            qchem_command="qchem -slurm",
            max_cores=32,
            input_file="mol.qin",
            output_file="mol.qout",
            linked=False,
        )
        expected_next = QCJob(
            qchem_command="qchem -slurm",
            max_cores=32,
            multimode="openmp",
            input_file="mol.qin",
            output_file="mol.qout",
            suffix=".opt_0",
            backup=True,
        ).as_dict()
        assert next(job).as_dict() == expected_next
        expected_next = QCJob(
            qchem_command="qchem -slurm",
            max_cores=32,
            multimode="openmp",
            input_file="mol.qin",
            output_file="mol.qout",
            suffix=".freq_0",
            backup=False,
        ).as_dict()
        assert next(job).as_dict() == expected_next
        assert (
            QCInput.from_file(f"{TEST_DIR}/5690_frag18/mol.qin.freq_0").as_dict()
            == QCInput.from_file(os.path.join(SCR_DIR, "mol.qin")).as_dict()
        )
        expected_next = QCJob(
            qchem_command="qchem -slurm",
            max_cores=32,
            multimode="openmp",
            input_file="mol.qin",
            output_file="mol.qout",
            suffix=".opt_1",
            backup=False,
        ).as_dict()
        assert next(job).as_dict() == expected_next
        assert (
            QCInput.from_file(f"{TEST_DIR}/5690_frag18/mol.qin.opt_1").as_dict()
            == QCInput.from_file(os.path.join(SCR_DIR, "mol.qin")).as_dict()
        )
        expected_next = QCJob(
            qchem_command="qchem -slurm",
            max_cores=32,
            multimode="openmp",
            input_file="mol.qin",
            output_file="mol.qout",
            suffix=".freq_1",
            backup=False,
        ).as_dict()
        assert next(job).as_dict() == expected_next
        assert (
            QCInput.from_file(f"{TEST_DIR}/5690_frag18/mol.qin.freq_1").as_dict()
            == QCInput.from_file(os.path.join(SCR_DIR, "mol.qin")).as_dict()
        )
        expected_next = QCJob(
            qchem_command="qchem -slurm",
            max_cores=32,
            multimode="openmp",
            input_file="mol.qin",
            output_file="mol.qout",
            suffix=".opt_2",
            backup=False,
        ).as_dict()
        assert next(job).as_dict() == expected_next
        assert (
            QCInput.from_file(f"{TEST_DIR}/5690_frag18/mol.qin.opt_2").as_dict()
            == QCInput.from_file(os.path.join(SCR_DIR, "mol.qin")).as_dict()
        )
        expected_next = QCJob(
            qchem_command="qchem -slurm",
            max_cores=32,
            multimode="openmp",
            input_file="mol.qin",
            output_file="mol.qout",
            suffix=".freq_2",
            backup=False,
        ).as_dict()
        assert next(job).as_dict() == expected_next
        assert (
            QCInput.from_file(f"{TEST_DIR}/5690_frag18/mol.qin.freq_2").as_dict()
            == QCInput.from_file(os.path.join(SCR_DIR, "mol.qin")).as_dict()
        )
        with pytest.raises(ValueError, match="ERROR: Can't deal with multiple neg frequencies yet! Exiting..."):
            next(job)


@skip_if_no_openbabel
class OptFFSmallNegFreqTest(TestCase):
    def setUp(self) -> None:
        os.makedirs(f"{SCR_DIR}/scratch", exist_ok=True)
        shutil.copyfile(f"{TEST_DIR}/small_neg_freq/mol.qin.orig", f"{SCR_DIR}/mol.qin")
        shutil.copyfile(f"{TEST_DIR}/small_neg_freq/mol.qin.opt_0", f"{SCR_DIR}/mol.qin.opt_0")
        shutil.copyfile(f"{TEST_DIR}/small_neg_freq/mol.qout.opt_0", f"{SCR_DIR}/mol.qout.opt_0")
        shutil.copyfile(f"{TEST_DIR}/small_neg_freq/mol.qout.freq_0", f"{SCR_DIR}/mol.qout.freq_0")
        shutil.copyfile(f"{TEST_DIR}/small_neg_freq/mol.qout.opt_1", f"{SCR_DIR}/mol.qout.opt_1")
        shutil.copyfile(f"{TEST_DIR}/small_neg_freq/mol.qout.freq_1", f"{SCR_DIR}/mol.qout.freq_1")
        shutil.copyfile(f"{TEST_DIR}/small_neg_freq/mol.qout.opt_2", f"{SCR_DIR}/mol.qout.opt_2")
        shutil.copyfile(f"{TEST_DIR}/small_neg_freq/mol.qout.freq_2", f"{SCR_DIR}/mol.qout.freq_2")
        os.chdir(SCR_DIR)

    def tearDown(self) -> None:
        os.chdir(CWD)
        shutil.rmtree(SCR_DIR)

    def test_OptFF(self) -> None:
        job = QCJob.opt_with_frequency_flattener(
            qchem_command="qchem -slurm",
            max_cores=32,
            input_file="mol.qin",
            output_file="mol.qout",
            linked=True,
        )
        expected_next = QCJob(
            qchem_command="qchem -slurm",
            max_cores=32,
            multimode="openmp",
            input_file="mol.qin",
            output_file="mol.qout",
            suffix=".opt_0",
            save_scratch=True,
            backup=True,
        ).as_dict()
        assert next(job).as_dict() == expected_next
        expected_next = QCJob(
            qchem_command="qchem -slurm",
            max_cores=32,
            multimode="openmp",
            input_file="mol.qin",
            output_file="mol.qout",
            suffix=".freq_0",
            save_scratch=True,
            backup=False,
        ).as_dict()
        assert next(job).as_dict() == expected_next
        assert (
            QCInput.from_file(f"{TEST_DIR}/small_neg_freq/mol.qin.freq_0").as_dict()
            == QCInput.from_file(os.path.join(SCR_DIR, "mol.qin")).as_dict()
        )
        shutil.copyfile(
            os.path.join(SCR_DIR, "mol.qin"),
            os.path.join(SCR_DIR, "mol.qin.freq_0"),
        )
        expected_next = QCJob(
            qchem_command="qchem -slurm",
            max_cores=32,
            multimode="openmp",
            input_file="mol.qin",
            output_file="mol.qout",
            suffix=".opt_1",
            save_scratch=True,
            backup=False,
        ).as_dict()
        assert next(job).as_dict() == expected_next
        assert (
            QCInput.from_file(f"{TEST_DIR}/small_neg_freq/mol.qin.opt_1").as_dict()
            == QCInput.from_file(os.path.join(SCR_DIR, "mol.qin")).as_dict()
        )
        shutil.copyfile(
            os.path.join(SCR_DIR, "mol.qin"),
            os.path.join(SCR_DIR, "mol.qin.opt_1"),
        )
        expected_next = QCJob(
            qchem_command="qchem -slurm",
            max_cores=32,
            multimode="openmp",
            input_file="mol.qin",
            output_file="mol.qout",
            suffix=".freq_1",
            save_scratch=True,
            backup=False,
        ).as_dict()
        assert next(job).as_dict() == expected_next
        assert (
            QCInput.from_file(f"{TEST_DIR}/small_neg_freq/mol.qin.freq_1").as_dict()
            == QCInput.from_file(os.path.join(SCR_DIR, "mol.qin")).as_dict()
        )
        shutil.copyfile(
            os.path.join(SCR_DIR, "mol.qin"),
            os.path.join(SCR_DIR, "mol.qin.freq_1"),
        )
        expected_next = QCJob(
            qchem_command="qchem -slurm",
            max_cores=32,
            multimode="openmp",
            input_file="mol.qin",
            output_file="mol.qout",
            suffix=".opt_2",
            save_scratch=True,
            backup=False,
        ).as_dict()
        assert next(job).as_dict() == expected_next
        assert (
            QCInput.from_file(f"{TEST_DIR}/small_neg_freq/mol.qin.opt_2").as_dict()
            == QCInput.from_file(os.path.join(SCR_DIR, "mol.qin")).as_dict()
        )
        shutil.copyfile(
            os.path.join(SCR_DIR, "mol.qin"),
            os.path.join(SCR_DIR, "mol.qin.opt_2"),
        )
        expected_next = QCJob(
            qchem_command="qchem -slurm",
            max_cores=32,
            multimode="openmp",
            input_file="mol.qin",
            output_file="mol.qout",
            suffix=".freq_2",
            save_scratch=True,
            backup=False,
        ).as_dict()
        assert next(job).as_dict() == expected_next
        assert (
            QCInput.from_file(f"{TEST_DIR}/small_neg_freq/mol.qin.freq_2").as_dict()
            == QCInput.from_file(os.path.join(SCR_DIR, "mol.qin")).as_dict()
        )
        shutil.copyfile(
            os.path.join(SCR_DIR, "mol.qin"),
            os.path.join(SCR_DIR, "mol.qin.freq_2"),
        )
        with pytest.raises(StopIteration):
            next(job)


@skip_if_no_openbabel
class OptFFSingleFreqFragsTest(TestCase):
    def setUp(self) -> None:
        os.makedirs(f"{SCR_DIR}/scratch", exist_ok=True)
        shutil.copyfile(f"{TEST_DIR}/single_freq_frag/mol.qin.orig", f"{SCR_DIR}/mol.qin")
        shutil.copyfile(f"{TEST_DIR}/single_freq_frag/mol.qin.opt_0", f"{SCR_DIR}/mol.qin.opt_0")
        shutil.copyfile(f"{TEST_DIR}/single_freq_frag/mol.qout.opt_0", f"{SCR_DIR}/mol.qout.opt_0")
        shutil.copyfile(f"{TEST_DIR}/single_freq_frag/mol.qin.freq_0", f"{SCR_DIR}/mol.qin.freq_0")
        shutil.copyfile(f"{TEST_DIR}/single_freq_frag/mol.qout.freq_0", f"{SCR_DIR}/mol.qout.freq_0")

        os.chdir(SCR_DIR)

    def tearDown(self) -> None:
        os.chdir(CWD)
        shutil.rmtree(SCR_DIR)

    def test_OptFF(self) -> None:
        job = QCJob.opt_with_frequency_flattener(
            qchem_command="qchem -slurm",
            max_cores=32,
            input_file="mol.qin",
            output_file="mol.qout",
            linked=True,
        )
        expected_next = QCJob(
            qchem_command="qchem -slurm",
            max_cores=32,
            multimode="openmp",
            input_file="mol.qin",
            output_file="mol.qout",
            suffix=".opt_0",
            save_scratch=True,
            backup=True,
        ).as_dict()
        assert next(job).as_dict() == expected_next
        expected_next = QCJob(
            qchem_command="qchem -slurm",
            max_cores=32,
            multimode="openmp",
            input_file="mol.qin",
            output_file="mol.qout",
            suffix=".freq_0",
            save_scratch=True,
            backup=False,
        ).as_dict()
        assert next(job).as_dict() == expected_next
        assert (
            QCInput.from_file(f"{TEST_DIR}/single_freq_frag/mol.qin.freq_0").as_dict()
            == QCInput.from_file(f"{SCR_DIR}/mol.qin").as_dict()
        )
        shutil.copyfile(f"{SCR_DIR}/mol.qin", f"{SCR_DIR}/mol.qin.freq_0")

        with pytest.raises(StopIteration):
            next(job)


@skip_if_no_openbabel
class TSFFTest(TestCase):
    def setUp(self) -> None:
        os.makedirs(SCR_DIR)
        shutil.copyfile(f"{TEST_DIR}/fftsopt_no_freqfirst/mol.qin.freq_0", f"{SCR_DIR}/test.qin")
        shutil.copyfile(f"{TEST_DIR}/fftsopt_no_freqfirst/mol.qout.ts_0", f"{SCR_DIR}/test.qout.ts_0")
        shutil.copyfile(f"{TEST_DIR}/fftsopt_no_freqfirst/mol.qout.freq_0", f"{SCR_DIR}/test.qout.freq_0")
        shutil.copyfile(f"{TEST_DIR}/fftsopt_no_freqfirst/mol.qin.freq_0", f"{SCR_DIR}/test.qin.freq_0")
        os.chdir(SCR_DIR)

    def tearDown(self) -> None:
        os.chdir(CWD)
        shutil.rmtree(SCR_DIR)

    def test_OptFF(self) -> None:
        job = QCJob.opt_with_frequency_flattener(
            qchem_command="qchem",
            max_cores=32,
            input_file="test.qin",
            output_file="test.qout",
            linked=False,
            transition_state=True,
        )
        expected_next = QCJob(
            qchem_command="qchem",
            max_cores=32,
            multimode="openmp",
            input_file="test.qin",
            output_file="test.qout",
            suffix=".ts_0",
            backup=True,
        ).as_dict()
        assert next(job).as_dict() == expected_next
        expected_next = QCJob(
            qchem_command="qchem",
            max_cores=32,
            multimode="openmp",
            input_file="test.qin",
            output_file="test.qout",
            suffix=".freq_0",
            backup=False,
        ).as_dict()
        assert next(job).as_dict() == expected_next
        assert (
            QCInput.from_file(f"{TEST_DIR}/fftsopt_no_freqfirst/mol.qin.freq_0").as_dict()
            == QCInput.from_file(f"{SCR_DIR}/test.qin").as_dict()
        )
        with pytest.raises(StopIteration):
            next(job)


@skip_if_no_openbabel
class TSFFFreqfirstTest(TestCase):
    def setUp(self) -> None:
        os.makedirs(f"{SCR_DIR}/scratch", exist_ok=True)
        shutil.copyfile(f"{TEST_DIR}/fftsopt_freqfirst/mol.qin.orig", f"{SCR_DIR}/mol.qin")
        shutil.copyfile(f"{TEST_DIR}/fftsopt_freqfirst/mol.qin.freq_pre", f"{SCR_DIR}/mol.qin.freq_pre")
        shutil.copyfile(f"{TEST_DIR}/fftsopt_freqfirst/mol.qout.freq_pre", f"{SCR_DIR}/mol.qout.freq_pre")
        shutil.copyfile(f"{TEST_DIR}/fftsopt_freqfirst/mol.qout.ts_0", f"{SCR_DIR}/mol.qout.ts_0")
        shutil.copyfile(f"{TEST_DIR}/fftsopt_freqfirst/mol.qout.freq_0", f"{SCR_DIR}/mol.qout.freq_0")
        os.chdir(SCR_DIR)

    def tearDown(self) -> None:
        os.chdir(CWD)
        shutil.rmtree(SCR_DIR)

    def test_OptFF(self) -> None:
        job = QCJob.opt_with_frequency_flattener(
            qchem_command="qchem -slurm",
            max_cores=32,
            input_file="mol.qin",
            output_file="mol.qout",
            linked=True,
            transition_state=True,
            freq_before_opt=True,
        )
        expected_next = QCJob(
            qchem_command="qchem -slurm",
            max_cores=32,
            multimode="openmp",
            input_file="mol.qin",
            output_file="mol.qout",
            suffix=".freq_pre",
            save_scratch=True,
            backup=True,
        ).as_dict()
        assert next(job).as_dict() == expected_next
        expected_next = QCJob(
            qchem_command="qchem -slurm",
            max_cores=32,
            multimode="openmp",
            input_file="mol.qin",
            output_file="mol.qout",
            suffix=".ts_0",
            save_scratch=True,
            backup=False,
        ).as_dict()
        assert next(job).as_dict() == expected_next
        assert (
            QCInput.from_file(f"{TEST_DIR}/fftsopt_freqfirst/mol.qin.ts_0").as_dict()
            == QCInput.from_file(f"{SCR_DIR}/mol.qin").as_dict()
        )
        shutil.copyfile(f"{SCR_DIR}/mol.qin", f"{SCR_DIR}/mol.qin.ts_0")
        expected_next = QCJob(
            qchem_command="qchem -slurm",
            max_cores=32,
            multimode="openmp",
            input_file="mol.qin",
            output_file="mol.qout",
            suffix=".freq_0",
            save_scratch=True,
            backup=False,
        ).as_dict()
        assert next(job).as_dict() == expected_next
        assert (
            QCInput.from_file(f"{TEST_DIR}/fftsopt_freqfirst/mol.qin.freq_0").as_dict()
            == QCInput.from_file(f"{SCR_DIR}/mol.qin").as_dict()
        )
        shutil.copyfile(f"{SCR_DIR}/mol.qin", f"{SCR_DIR}/mol.qin.freq_0")
        with pytest.raises(StopIteration):
            next(job)


@skip_if_no_openbabel
class TSFFFreqFirstMultipleCyclesTest(TestCase):
    def setUp(self) -> None:
        os.makedirs(f"{SCR_DIR}/scratch", exist_ok=True)
        shutil.copyfile(f"{TEST_DIR}/fftsopt_multiple_cycles/mol.qin.orig", f"{SCR_DIR}/mol.qin")
        shutil.copyfile(f"{TEST_DIR}/fftsopt_multiple_cycles/mol.qin.freq_pre", f"{SCR_DIR}/mol.qin.freq_pre")
        shutil.copyfile(f"{TEST_DIR}/fftsopt_multiple_cycles/mol.qout.freq_pre", f"{SCR_DIR}/mol.qout.freq_pre")
        shutil.copyfile(f"{TEST_DIR}/fftsopt_multiple_cycles/mol.qout.ts_0", f"{SCR_DIR}/mol.qout.ts_0")
        shutil.copyfile(f"{TEST_DIR}/fftsopt_multiple_cycles/mol.qout.freq_0", f"{SCR_DIR}/mol.qout.freq_0")
        shutil.copyfile(f"{TEST_DIR}/fftsopt_multiple_cycles/mol.qout.ts_1", f"{SCR_DIR}/mol.qout.ts_1")
        shutil.copyfile(f"{TEST_DIR}/fftsopt_multiple_cycles/mol.qout.freq_1", f"{SCR_DIR}/mol.qout.freq_1")
        os.chdir(SCR_DIR)

    def tearDown(self) -> None:
        os.chdir(CWD)
        shutil.rmtree(SCR_DIR)

    def test_OptFF(self) -> None:
        job = QCJob.opt_with_frequency_flattener(
            qchem_command="qchem -slurm",
            max_cores=32,
            input_file="mol.qin",
            output_file="mol.qout",
            linked=True,
            transition_state=True,
            freq_before_opt=True,
        )
        expected_next = QCJob(
            qchem_command="qchem -slurm",
            max_cores=32,
            multimode="openmp",
            input_file="mol.qin",
            output_file="mol.qout",
            suffix=".freq_pre",
            save_scratch=True,
            backup=True,
        ).as_dict()
        assert next(job).as_dict() == expected_next
        expected_next = QCJob(
            qchem_command="qchem -slurm",
            max_cores=32,
            multimode="openmp",
            input_file="mol.qin",
            output_file="mol.qout",
            suffix=".ts_0",
            save_scratch=True,
            backup=False,
        ).as_dict()
        assert next(job).as_dict() == expected_next
        assert (
            QCInput.from_file(f"{TEST_DIR}/fftsopt_multiple_cycles/mol.qin.ts_0").as_dict()
            == QCInput.from_file(f"{SCR_DIR}/mol.qin").as_dict()
        )
        shutil.copyfile(f"{SCR_DIR}/mol.qin", f"{SCR_DIR}/mol.qin.ts_0")
        expected_next = QCJob(
            qchem_command="qchem -slurm",
            max_cores=32,
            multimode="openmp",
            input_file="mol.qin",
            output_file="mol.qout",
            suffix=".freq_0",
            save_scratch=True,
            backup=False,
        ).as_dict()
        assert next(job).as_dict() == expected_next
        assert (
            QCInput.from_file(f"{TEST_DIR}/fftsopt_multiple_cycles/mol.qin.freq_0").as_dict()
            == QCInput.from_file(f"{SCR_DIR}/mol.qin").as_dict()
        )
        shutil.copyfile(f"{SCR_DIR}/mol.qin", f"{SCR_DIR}/mol.qin.freq_0")
        expected_next = QCJob(
            qchem_command="qchem -slurm",
            max_cores=32,
            multimode="openmp",
            input_file="mol.qin",
            output_file="mol.qout",
            suffix=".ts_1",
            save_scratch=True,
            backup=False,
        ).as_dict()
        assert next(job).as_dict() == expected_next
        assert (
            QCInput.from_file(f"{TEST_DIR}/fftsopt_multiple_cycles/mol.qin.ts_1").as_dict()
            == QCInput.from_file(f"{SCR_DIR}/mol.qin").as_dict()
        )
        shutil.copyfile(f"{SCR_DIR}/mol.qin", f"{SCR_DIR}/mol.qin.ts_1")
        expected_next = QCJob(
            qchem_command="qchem -slurm",
            max_cores=32,
            multimode="openmp",
            input_file="mol.qin",
            output_file="mol.qout",
            suffix=".freq_1",
            save_scratch=True,
            backup=False,
        ).as_dict()
        assert next(job).as_dict() == expected_next
        assert (
            QCInput.from_file(f"{TEST_DIR}/fftsopt_multiple_cycles/mol.qin.freq_1").as_dict()
            == QCInput.from_file(f"{SCR_DIR}/mol.qin").as_dict()
        )
        shutil.copyfile(f"{SCR_DIR}/mol.qin", f"{SCR_DIR}/mol.qin.freq_1")

        with pytest.raises(StopIteration):
            next(job)



"""Utility methods for VASP error handlers."""

from __future__ import annotations

from typing import TYPE_CHECKING, Any

import numpy as np
from pymatgen.io.vasp.inputs import Kpoints

if TYPE_CHECKING:
    from pymatgen.core import Structure


def _estimate_num_k_points_from_kspacing(structure: Structure, kspacing: float) -> tuple[int, ...]:
    """
    Estimate the number of k-points used by VASP.

    Inputs:
        structure (Structure): structure used in the calculation
        kspacing (float): KSPACING used in the calculation

    Returns:
        tuple[int,int,int] : the number of estimated k-points on each axis.

    Note that there is a typo in the VASP manual:
        https://www.vasp.at/wiki/index.php/KSPACING
    The inner product between direct a_i and reciprocal b_i
    lattice vectors is conventionally a_i . b_j = 2 pi delta_ij

    That leads to an extra 2 pi factor in the expression for the
    number of k-points per axis from KSPACING. The formula used
    below has been checked for accuracy against VASP calculations.
    """
    return tuple(int(max(1, np.ceil(structure.lattice.reciprocal_lattice.abc[i] / kspacing))) for i in range(3))


def increase_k_point_density(
    kpoints: Kpoints | dict | float,
    structure: Structure,
    factor: float = 0.1,
    max_inc: int = 500,
    min_kpoints: int = 1,
    force_gamma: bool = True,
) -> dict:
    """
    Inputs:
        kpoints (Kpoints, dict, float, int) :
            If a dict or Kpoints object: original Kpoints used in the calculation
            If a float: the original KSPACING used in the calculation
        structure (Structure) : associated structure
        factor (float) : factor used to increase k-point density.
            The first increase uses approximately (1 + factor) higher k-point density.
            The second increase: ~ (1 + 2*factor) higher k-kpoint density, etc.
        max_inc (int) : the maximum permitted increases in k-point density
            before giving up
        min_kpoints (int): The minimum permitted number of k-points.
            Can be useful if using the tetrahedron method, where
            at least 4 k-points are needed.
        force_gamma (bool) = True: whether to use Gamma-centered or
            Monkhorst-Pack grids
    Outputs:
        dict :
            The new Kpoints object / KSPACING consistent with constraints.
            If an empty dict, no new k-point mesh could be found.
    """
    uses_kspacing = isinstance(kpoints, float | int)

    if uses_kspacing:
        orig_num_kpoints = np.prod(_estimate_num_k_points_from_kspacing(structure, kpoints))  # type: ignore[arg-type]

    else:
        if isinstance(kpoints, Kpoints):
            kpoints = kpoints.as_dict()

        orig_num_kpoints = np.prod(kpoints["kpoints"][0])  # type: ignore[index]

        # try to approximate k-points per reciprocal atom used in pymatgen
        lengths = structure.lattice.abc
        mult = max([nk * lengths[ik] for ik, nk in enumerate(kpoints["kpoints"][0])])  # type: ignore[index]
        ngrid = mult**3 / np.prod(lengths)
        kppa = len(structure) * ngrid

    mult_fac = 1.0 + factor
    min_kpoints = max(min_kpoints, 1)

    success = False
    for _ in range(max_inc):
        if uses_kspacing:
            new_kpoints: dict[str, Any] = {"KSPACING": round(kpoints / mult_fac, 6), "KGAMMA": force_gamma}  # type: ignore[operator]
            new_nk = _estimate_num_k_points_from_kspacing(structure, new_kpoints["KSPACING"])
        else:
            kpts = Kpoints.automatic_density(structure, mult_fac * kppa, force_gamma=force_gamma)  # type: ignore
            new_kpoints = {
                "generation_style": str(kpts.style),
                "kpoints": (tuple(kpts.kpts[0]),),
            }
            new_nk = new_kpoints["kpoints"][0]  # type: ignore[index]

        if (new_num_kpoints := np.prod(new_nk)) > orig_num_kpoints and (new_num_kpoints >= min_kpoints):
            success = True
            break

        mult_fac += factor

    return new_kpoints if success else {}  # type: ignore



"""Created on Jun 1, 2012."""

import datetime
import os
import shutil
import tarfile
from glob import glob
from pathlib import Path

import pytest
from monty.io import zopen
from monty.os.path import zpath
from pymatgen.io.vasp.inputs import Incar, Kpoints, Structure, VaspInput
from pymatgen.util.testing import MatSciTest

from custodian.utils import tracked_lru_cache
from custodian.vasp.handlers import (
    AliasingErrorHandler,
    DriftErrorHandler,
    FrozenJobErrorHandler,
    IncorrectSmearingHandler,
    KspacingMetalHandler,
    LargeSigmaHandler,
    LrfCommutatorHandler,
    MeshSymmetryErrorHandler,
    NonConvergingErrorHandler,
    PositiveEnergyErrorHandler,
    PotimErrorHandler,
    ScanMetalHandler,
    StdErrHandler,
    UnconvergedErrorHandler,
    VaspErrorHandler,
    WalltimeHandler,
)
from custodian.vasp.interpreter import VaspModder
from tests.conftest import TEST_FILES

__author__ = "Shyue Ping Ong, Stephen Dacek, Janosh Riebesell"
__copyright__ = "Copyright 2012, The Materials Project"
__version__ = "0.1"
__maintainer__ = "Shyue Ping Ong"
__email__ = "shyue@mit.edu"
__date__ = "Jun 1, 2012"

CWD = os.getcwd()
os.environ.setdefault("PMG_VASP_PSP_DIR", TEST_FILES)


@pytest.fixture(autouse=True)
def _clear_tracked_cache() -> None:
    """Clear the cache of the stored functions between the tests."""

    tracked_lru_cache.tracked_cache_clear()


def copy_tmp_files(tmp_path: str, *file_paths: str) -> None:
    for file_path in file_paths:
        for ext in (".gz", ".bz2"):
            # TODO remove this when https://github.com/materialsvirtuallab/monty/pull/682
            # is released
            file_path = file_path.removesuffix(ext)
        src_path = zpath(f"{TEST_FILES}/{file_path}")
        dst_path = f"{tmp_path}/{os.path.basename(src_path)}"
        if os.path.isdir(src_path):
            shutil.copytree(src_path, dst_path)
        else:
            shutil.copy(src_path, dst_path)
    os.chdir(tmp_path)


class VaspErrorHandlerTest(MatSciTest):
    def setUp(self) -> None:
        copy_tmp_files(self.tmp_path, *glob("*", root_dir=TEST_FILES))

    def test_frozen_job(self) -> None:
        handler = FrozenJobErrorHandler()
        dct = handler.correct()
        assert dct["errors"] == ["Frozen job"]
        assert Incar.from_file("INCAR")["ALGO"] == "Normal"

    def test_algotet(self) -> None:
        shutil.copy("INCAR.algo_tet_only", "INCAR")
        handler = VaspErrorHandler("vasp.algo_tet_only")
        handler.check()
        dct = handler.correct()
        assert dct["errors"] == ["algo_tet"]
        assert dct["actions"] == [{"action": {"_set": {"ALGO": "Fast"}}, "dict": "INCAR"}]
        assert handler.error_count["algo_tet"] == 1

        # 2nd error should set ISMEAR to 0.
        handler.check()
        dct = handler.correct()
        assert dct["errors"] == ["algo_tet"]
        assert handler.error_count["algo_tet"] == 2
        assert dct["actions"] == [{"action": {"_set": {"ISMEAR": 0, "SIGMA": 0.05}}, "dict": "INCAR"}]

    def test_subspace(self) -> None:
        handler = VaspErrorHandler("vasp.subspace")
        handler.check()
        dct = handler.correct()
        assert dct["errors"] == ["subspacematrix"]
        assert dct["actions"] == [{"action": {"_set": {"LREAL": False}}, "dict": "INCAR"}]

        # 2nd error should set PREC to accurate.
        handler.check()
        dct = handler.correct()
        assert dct["errors"] == ["subspacematrix"]
        assert dct["actions"] == [{"action": {"_set": {"PREC": "Accurate"}}, "dict": "INCAR"}]

    def test_check_correct(self) -> None:
        handler = VaspErrorHandler("vasp.ksymm")
        handler.check()
        dct = handler.correct()
        assert dct["errors"] == ["ksymm"]
        assert dct["actions"] == [{"action": {"_set": {"SYMPREC": 1.0e-4}}, "dict": "INCAR"}]

        handler = VaspErrorHandler("vasp.ksymm", errors_subset_to_catch=["eddrmm"])
        assert not handler.check()

        handler = VaspErrorHandler("vasp.sgrcon")
        handler.check()
        dct = handler.correct()
        assert dct["errors"] == ["rot_matrix"]
        assert {a["dict"] for a in dct["actions"]} == {"KPOINTS"}

        handler = VaspErrorHandler("vasp.real_optlay")
        handler.check()
        dct = handler.correct()
        assert dct["errors"] == ["real_optlay"]
        assert dct["actions"] == [{"action": {"_set": {"LREAL": False}}, "dict": "INCAR"}]

    def test_mesh_symmetry(self) -> None:
        handler = MeshSymmetryErrorHandler("vasp.ibzkpt")
        handler.check()
        dct = handler.correct()
        assert dct["errors"] == ["mesh_symmetry"]
        assert dct["actions"] == [{"action": {"_set": {"kpoints": [[4, 4, 4]]}}, "dict": "KPOINTS"}]

    def test_brions(self) -> None:
        shutil.copy("INCAR.ibrion", "INCAR")
        handler = VaspErrorHandler("vasp.brions")
        handler.check()
        dct = handler.correct()
        assert dct["errors"] == ["brions"]
        incar = Incar.from_file("INCAR")
        assert incar["IBRION"] == 1
        assert incar["POTIM"] == pytest.approx(1.5)

        handler.check()
        dct = handler.correct()
        assert dct["errors"] == ["brions"]
        incar = Incar.from_file("INCAR")
        assert incar["IBRION"] == 2
        assert incar["POTIM"] == pytest.approx(0.5)

    def test_dentet(self) -> None:
        # ensure we use a very small, Monkhorst-Pack k-point grid to check all handler functionality
        Kpoints(style="Monkhorst").write_file("KPOINTS")

        handler = VaspErrorHandler("vasp.dentet")
        handler.check()
        dct = handler.correct()
        assert dct["errors"] == ["dentet"]
        assert dct["actions"] == [{"action": {"_set": {"generation_style": "Gamma"}}, "dict": "KPOINTS"}]

        handler.check()
        dct = handler.correct()
        assert dct["errors"] == ["dentet"]
        assert dct["actions"] == [
            {"action": {"_set": {"kpoints": ((3, 2, 1),), "generation_style": "Gamma"}}, "dict": "KPOINTS"}
        ]

        handler.check()
        dct = handler.correct()
        assert dct["errors"] == ["dentet"]
        assert dct["actions"] == [{"action": {"_set": {"ISMEAR": 0, "SIGMA": 0.05}}, "dict": "INCAR"}]

    def test_zbrent(self) -> None:
        handler = VaspErrorHandler("vasp.zbrent")
        handler.check()
        dct = handler.correct()
        assert dct["errors"] == ["zbrent"]
        incar = Incar.from_file("INCAR")
        assert incar["IBRION"] == 2
        assert incar["EDIFF"] == 1e-06
        assert incar["NELMIN"] == 8

        handler.check()
        dct = handler.correct()
        assert dct["errors"] == ["zbrent"]
        incar = Incar.from_file("INCAR")
        assert incar["IBRION"] == 1
        assert incar["EDIFF"] == 1e-07
        assert incar["NELMIN"] == 8

        shutil.copy(f"{TEST_FILES}/INCAR", f"{self.tmp_path}/INCAR")
        handler = VaspErrorHandler("vasp.zbrent")
        handler.vtst_fixes = True
        handler.check()
        dct = handler.correct()
        assert dct["errors"] == ["zbrent"]
        incar = Incar.from_file("INCAR")
        assert incar["IBRION"] == 3
        assert incar["IOPT"] == 7
        assert incar["POTIM"] == 0
        assert incar["EDIFF"] == 1e-06
        assert incar["NELMIN"] == 8

        handler.check()
        dct = handler.correct()
        assert dct["errors"] == ["zbrent"]
        incar = Incar.from_file("INCAR")
        assert incar["IBRION"] == 3
        assert incar["IOPT"] == 7
        assert incar["POTIM"] == 0
        assert incar["EDIFF"] == 1e-07
        assert incar["NELMIN"] == 8

        shutil.copy("INCAR.ediff", "INCAR")
        handler = VaspErrorHandler("vasp.zbrent")
        handler.check()
        dct = handler.correct()
        assert dct["errors"] == ["zbrent"]
        incar = Incar.from_file("INCAR")
        assert incar["IBRION"] == 2
        assert incar["EDIFF"] == 1e-07
        assert incar["NELMIN"] == 8

        handler.check()
        dct = handler.correct()
        assert dct["errors"] == ["zbrent"]
        incar = Incar.from_file("INCAR")
        assert incar["IBRION"] == 1
        assert incar["EDIFF"] == 1e-08
        assert incar["NELMIN"] == 8

    def test_brmix(self) -> None:
        handler = VaspErrorHandler("vasp.brmix")
        assert handler.check() is True

        # The first (no good OUTCAR) correction, check IMIX
        dct = handler.correct()
        assert dct["errors"] == ["brmix"]
        vi = VaspInput.from_directory(".")
        assert vi["INCAR"]["IMIX"] == 1
        assert os.path.isfile("CHGCAR")

        # The next correction check Gamma and evenize
        handler.correct()
        vi = VaspInput.from_directory(".")
        assert "IMIX" not in vi["INCAR"]
        assert os.path.isfile("CHGCAR")
        if vi["KPOINTS"].style == Kpoints.supported_modes.Gamma and vi["KPOINTS"].num_kpts < 1:
            all_kpts_even = all(n % 2 == 0 for n in vi["KPOINTS"].kpts[0])
            assert not all_kpts_even

        # The next correction check ISYM and no CHGCAR
        handler.correct()
        vi = VaspInput.from_directory(".")
        assert vi["INCAR"]["ISYM"] == 0
        assert not os.path.isfile("CHGCAR")

        shutil.copy("INCAR.nelect", "INCAR")
        handler = VaspErrorHandler("vasp.brmix")
        assert handler.check() is False
        dct = handler.correct()
        assert dct["errors"] == []

    def test_too_few_bands(self) -> None:
        shutil.copytree(f"{TEST_FILES}/too_few_bands", self.tmp_path, dirs_exist_ok=True, symlinks=True)
        os.chdir(self.tmp_path)
        shutil.copy("INCAR", "INCAR.orig")
        handler = VaspErrorHandler("vasp.too_few_bands")
        handler.check()
        dct = handler.correct()
        assert dct["errors"] == ["too_few_bands"]
        assert dct["actions"] == [{"action": {"_set": {"NBANDS": 501}}, "dict": "INCAR"}]

    def test_rot_matrix(self) -> None:
        shutil.copytree(f"{TEST_FILES}/poscar_error", self.tmp_path, dirs_exist_ok=True, symlinks=True)
        os.chdir(self.tmp_path)
        shutil.copy("KPOINTS", "KPOINTS.orig")
        handler = VaspErrorHandler()
        handler.check()
        dct = handler.correct()
        assert dct["errors"] == ["rot_matrix"]

    def test_rot_matrix_vasp6(self) -> None:
        handler = VaspErrorHandler("vasp6.sgrcon")
        assert handler.check() is True
        assert handler.correct()["errors"] == ["rot_matrix"]

    def test_coef(self) -> None:
        handler = VaspErrorHandler("vasp6.coef")
        handler.check()
        dct = handler.correct()
        assert dct["actions"] == [{"file": "WAVECAR", "action": {"_file_delete": {"mode": "actual"}}}]

        handler = VaspErrorHandler("vasp6.coef2")
        handler.check()
        dct = handler.correct()
        assert dct["actions"] == [{"file": "WAVECAR", "action": {"_file_delete": {"mode": "actual"}}}]

    def test_as_from_dict(self) -> None:
        handler = VaspErrorHandler("random_name")
        h2 = VaspErrorHandler.from_dict(handler.as_dict())
        assert isinstance(h2, VaspErrorHandler)
        assert h2.output_filename == "random_name"

    def test_pssyevx_pdsyevx(self) -> None:
        incar_orig = Incar.from_file("INCAR")
        # Joining tests for these three tags as they have identical handling
        for error_name in ("pssyevx", "pdsyevx"):
            handler = VaspErrorHandler(f"vasp.{error_name}")
            assert handler.check() is True
            assert handler.correct()["errors"] == [error_name]
            incar = Incar.from_file("INCAR")
            assert incar["ALGO"] == "Normal"
            incar_orig.write_file("INCAR")

    def test_eddrmm(self) -> None:
        shutil.copy("CONTCAR.eddav_eddrmm", "CONTCAR")
        handler = VaspErrorHandler("vasp.eddrmm")
        assert handler.check() is True
        assert handler.correct()["errors"] == ["eddrmm"]
        incar = Incar.from_file("INCAR")
        assert incar["ALGO"] == "Normal"
        assert handler.correct()["errors"] == ["eddrmm"]
        incar = Incar.from_file("INCAR")
        assert incar["POTIM"] == 0.25
        p = Structure.from_file("POSCAR")
        c = Structure.from_file("CONTCAR")
        assert p == c

    def test_nicht_konv(self) -> None:
        handler = VaspErrorHandler("vasp.nicht_konvergent")
        assert handler.check() is True
        assert handler.correct()["errors"] == ["nicht_konv"]
        incar = Incar.from_file("INCAR")
        assert incar["LREAL"] is False

    def test_edddav(self) -> None:
        shutil.copy("CONTCAR.eddav_eddrmm", "CONTCAR")
        handler = VaspErrorHandler("vasp.edddav2")
        assert handler.check() is True
        assert handler.correct()["errors"] == ["edddav"]
        incar = Incar.from_file("INCAR")
        assert incar["NCORE"] == 2
        p = Structure.from_file("POSCAR")
        c = Structure.from_file("CONTCAR")
        assert p == c

        handler = VaspErrorHandler("vasp.edddav")
        assert handler.check() is True
        assert handler.correct()["errors"] == ["edddav"]
        assert not os.path.isfile("CHGCAR")
        p = Structure.from_file("POSCAR")
        c = Structure.from_file("CONTCAR")
        assert p == c

    def test_gradient_not_orthogonal(self) -> None:
        handler = VaspErrorHandler("vasp.gradient_not_orthogonal")
        assert handler.check() is True
        assert "grad_not_orth" in handler.correct()["errors"]
        incar = Incar.from_file("INCAR")
        assert incar["ALGO"] == "Fast"

        shutil.copy("INCAR.gga_all", "INCAR")
        handler = VaspErrorHandler("vasp.gradient_not_orthogonal")
        assert handler.check() is True
        assert "grad_not_orth" in handler.correct()["errors"]
        incar = Incar.from_file("INCAR")
        assert incar["ALGO"] == "Fast"

        shutil.copy("INCAR.hybrid_normal", "INCAR")
        handler = VaspErrorHandler("vasp.gradient_not_orthogonal")
        assert handler.check() is True
        assert "grad_not_orth" in handler.correct()["errors"]
        incar = Incar.from_file("INCAR")
        assert incar["ALGO"] == "Normal"

        shutil.copy("INCAR.hybrid_all", "INCAR")
        handler = VaspErrorHandler("vasp.gradient_not_orthogonal")
        assert handler.check() is True
        assert "grad_not_orth" in handler.correct()["errors"]
        incar = Incar.from_file("INCAR")
        assert incar["ALGO"] == "Normal"

        shutil.copy("INCAR.metagga_all", "INCAR")
        handler = VaspErrorHandler("vasp.gradient_not_orthogonal")
        assert handler.check() is True
        assert "grad_not_orth" in handler.correct()["errors"]
        incar = Incar.from_file("INCAR")
        assert incar["ALGO"] == "Normal"

    def test_rhosyg(self) -> None:
        handler = VaspErrorHandler("vasp.rhosyg")
        assert handler.check() is True
        assert handler.correct()["errors"] == ["rhosyg"]
        incar = Incar.from_file("INCAR")
        assert incar["SYMPREC"] == 0.0001
        assert handler.correct()["errors"] == ["rhosyg"]
        incar = Incar.from_file("INCAR")
        assert incar["ISYM"] == 0

    def test_rhosyg_vasp6(self) -> None:
        handler = VaspErrorHandler("vasp6.rhosyg")
        assert handler.check() is True
        assert handler.correct()["errors"] == ["rhosyg"]
        incar = Incar.from_file("INCAR")
        assert incar["SYMPREC"] == 0.0001
        assert handler.correct()["errors"] == ["rhosyg"]
        incar = Incar.from_file("INCAR")
        assert incar["ISYM"] == 0

    def test_hnform(self) -> None:
        handler = VaspErrorHandler("vasp.hnform")
        assert handler.check() is True
        assert handler.correct()["errors"] == ["hnform"]
        incar = Incar.from_file("INCAR")
        assert incar["ISYM"] == 0

    def test_bravais(self) -> None:
        handler = VaspErrorHandler("vasp6.bravais")
        assert handler.check() is True
        assert handler.correct()["errors"] == ["bravais"]
        incar = Incar.from_file("INCAR")
        assert incar["SYMPREC"] == 0.0001

        shutil.copy("INCAR.symprec", "INCAR")
        handler = VaspErrorHandler("vasp6.bravais")
        assert handler.check() is True
        assert handler.correct()["errors"] == ["bravais"]
        incar = Incar.from_file("INCAR")
        assert incar["SYMPREC"] == 1e-6

    def test_posmap_and_pricelv(self) -> None:
        incar_orig = Incar.from_file("INCAR")
        # Joining tests for these three tags as they have identical handling
        for error_name in ("posmap", "posmap-6", "pricelv"):
            if error_name == "posmap-6":
                vasp_std_out = "vasp6.posmap"
                error_name = "posmap"
            else:
                vasp_std_out = f"vasp.{error_name}"

            handler = VaspErrorHandler(vasp_std_out)
            assert handler.check() is True
            assert handler.correct()["errors"] == [error_name]
            incar = Incar.from_file("INCAR")
            assert incar["SYMPREC"] == pytest.approx(1e-6)

            assert handler.check() is True
            assert handler.correct()["errors"] == [error_name]
            incar = Incar.from_file("INCAR")
            assert incar["SYMPREC"] == pytest.approx(1e-4)

            assert handler.check() is True
            assert handler.correct()["errors"] == [error_name]
            incar = Incar.from_file("INCAR")
            assert incar["ISYM"] == 0

            incar_orig.write_file("INCAR")

    def test_point_group(self) -> None:
        handler = VaspErrorHandler("vasp.point_group")
        assert handler.check() is True
        assert handler.correct()["errors"] == ["point_group"]
        incar = Incar.from_file("INCAR")
        assert incar["ISYM"] == 0

    def test_symprec_noise(self) -> None:
        handler = VaspErrorHandler("vasp.symprec_noise")
        assert handler.check() is True
        assert handler.correct()["errors"] == ["symprec_noise"]
        incar = Incar.from_file("INCAR")
        assert incar["SYMPREC"] == 1e-06

    def test_dfpt_ncore(self) -> None:
        handler = VaspErrorHandler("vasp.dfpt_ncore")
        assert handler.check() is True
        assert handler.correct()["errors"] == ["dfpt_ncore"]
        incar = Incar.from_file("INCAR")
        assert "NPAR" not in incar
        assert "NCORE" not in incar

    def test_finite_difference_ncore(self) -> None:
        handler = VaspErrorHandler("vasp.fd_ncore")
        assert handler.check() is True
        assert handler.correct()["errors"] == ["dfpt_ncore"]
        incar = Incar.from_file("INCAR")
        assert "NPAR" not in incar
        assert "NCORE" not in incar

    def test_point_group_vasp6(self) -> None:
        # the error message is formatted differently in VASP6 compared to VASP5
        handler = VaspErrorHandler("vasp6.point_group")
        assert handler.check() is True
        assert handler.correct()["errors"] == ["point_group"]
        incar = Incar.from_file("INCAR")
        assert incar["ISYM"] == 0

    def test_inv_rot_matrix_vasp6(self) -> None:
        # the error message is formatted differently in VASP6 compared to VASP5
        handler = VaspErrorHandler("vasp6.inv_rot_mat")
        assert handler.check() is True
        assert handler.correct()["errors"] == ["inv_rot_mat"]
        incar = Incar.from_file("INCAR")
        assert incar["SYMPREC"] == 1e-08

    def test_bzint_vasp6(self) -> None:
        # the BZINT error message is formatted differently in VASP6 compared to VASP5
        handler = VaspErrorHandler("vasp6.bzint")
        assert handler.check() is True
        dct = handler.correct()
        assert dct["errors"] == ["tet"]
        incar = Incar.from_file("INCAR")
        assert incar["ISMEAR"] == 0
        assert incar["SIGMA"] == 0.05

    def test_too_large_kspacing(self) -> None:
        shutil.copy("INCAR.kspacing", "INCAR")
        os.remove("KPOINTS")
        VaspInput.from_directory(".")
        handler = VaspErrorHandler("vasp.dentet")
        handler.check()
        dct = handler.correct()
        assert dct["errors"] == ["dentet"]
        assert dct["actions"] == [{"action": {"_set": {"KSPACING": 1.333333, "KGAMMA": True}}, "dict": "INCAR"}]

    def test_nbands_not_sufficient(self) -> None:
        handler = VaspErrorHandler("vasp.nbands_not_sufficient")
        shutil.copy("OUTCAR_auto_nbands", "OUTCAR")
        assert handler.check() is True
        dct = handler.correct()
        assert "nbands_not_sufficient" in dct["errors"]
        assert dct["actions"] == [{"action": {"_set": {"NBANDS": 9}}, "dict": "INCAR"}]

    def test_too_few_bands_round_error(self) -> None:
        # originally there are NBANDS= 7
        # correction should increase it
        shutil.copy("INCAR.too_few_bands_round_error", "INCAR")
        handler = VaspErrorHandler("vasp.too_few_bands_round_error")
        assert handler.check() is True
        dct = handler.correct()
        assert dct["errors"] == ["too_few_bands"]
        assert dct["actions"] == [{"dict": "INCAR", "action": {"_set": {"NBANDS": 8}}}]

    def test_set_core_wf(self) -> None:
        handler = VaspErrorHandler("vasp.set_core_wf")
        assert handler.check() is True
        dct = handler.correct()
        assert dct["errors"] == ["set_core_wf"]
        assert dct["actions"] is None

    def test_read_error(self) -> None:
        handler = VaspErrorHandler("vasp.read_error")
        assert handler.check() is True
        dct = handler.correct()
        assert dct["errors"] == ["read_error"]
        assert dct["actions"] is None

    def test_harris(self) -> None:
        handler = VaspErrorHandler("vasp.harris")
        assert handler.check() is True
        dct = handler.correct()
        assert dct["errors"] == ["spin_polarized_harris"]
        assert dct["actions"] is None

    def test_amin(self) -> None:
        # Cell with at least one dimension >= 50 A, but AMIN > 0.01, and calculation not yet complete
        shutil.copy("INCAR.amin", "INCAR")
        handler = VaspErrorHandler("vasp.amin")
        handler.check()
        dct = handler.correct()
        assert dct["errors"] == ["amin"]
        assert dct["actions"] == [{"action": {"_set": {"AMIN": 0.01}}, "dict": "INCAR"}]

    def test_eddiag(self) -> None:
        # subspace rotation error
        os.remove("CONTCAR")
        shutil.copy("INCAR.amin", "INCAR")
        handler = VaspErrorHandler("vasp.eddiag")
        handler.check()
        dct = handler.correct()
        assert dct["errors"] == ["eddiag"]
        # first check that no CONTCAR exists, only action should be updating INCAR
        # ALGO = Fast --> ALGO = Normal
        assert dct["actions"] == [{"action": {"_set": {"ALGO": "Normal"}}, "dict": "INCAR"}]

        # now copy CONTCAR and check that both CONTCAR->POSCAR
        # and INCAR updates are included: ALGO = Normal --> ALGO = exact
        shutil.copy("CONTCAR.eddiag", "CONTCAR")
        handler = VaspErrorHandler("vasp.eddiag")
        handler.check()
        dct = handler.correct()
        assert dct["actions"] == [
            {"file": "CONTCAR", "action": {"_file_copy": {"dest": "POSCAR"}}},
            {"action": {"_set": {"ALGO": "exact"}}, "dict": "INCAR"},
        ]

    def test_auto_nbands(self) -> None:
        shutil.copy("OUTCAR_auto_nbands", "OUTCAR")
        handler = VaspErrorHandler(zpath("vasp.auto_nbands"))
        handler.check()
        with pytest.warns(UserWarning, match="NBANDS seems to be too high"):
            dct = handler.correct()
        assert "auto_nbands" in dct["errors"]

    def test_auto_nbands_bad_parallelization(self) -> None:
        shutil.copy("OUTCAR_auto_nbands_parallel", "OUTCAR")
        # hacky way to deal with custodian CI unzipping test files
        Incar.from_file(zpath("INCAR.auto_nbands_parallel")).write_file("INCAR")

        handler = VaspErrorHandler(zpath("vasp.auto_nbands"))
        handler.check()
        with pytest.warns(UserWarning, match="setting was incompatible with your parallelization"):
            dct = handler.correct()
        assert "auto_nbands" in dct["errors"]
        assert dct["actions"] == [{"dict": "INCAR", "action": {"_set": {"NBANDS": 64}}}]


class AliasingErrorHandlerTest(MatSciTest):
    def setUp(self) -> None:
        copy_tmp_files(self.tmp_path, *glob("aliasing/*", root_dir=TEST_FILES))

    def test_aliasing(self) -> None:
        handler = AliasingErrorHandler("vasp.aliasing")
        handler.check()
        dct = handler.correct()

        assert dct["errors"] == ["aliasing"]
        assert dct["actions"] == [
            {"action": {"_set": {"NGX": 34}}, "dict": "INCAR"},
            {"file": "CHGCAR", "action": {"_file_delete": {"mode": "actual"}}},
            {"file": "WAVECAR", "action": {"_file_delete": {"mode": "actual"}}},
        ]

    def test_aliasing_incar(self) -> None:
        shutil.copy("INCAR", "INCAR.orig")
        handler = AliasingErrorHandler("vasp.aliasing_incar")
        handler.check()
        dct = handler.correct()

        assert dct["errors"] == ["aliasing_incar"]
        assert dct["actions"] == [
            {"action": {"_unset": {"NGY": 1, "NGZ": 1}}, "dict": "INCAR"},
            {"file": "CHGCAR", "action": {"_file_delete": {"mode": "actual"}}},
            {"file": "WAVECAR", "action": {"_file_delete": {"mode": "actual"}}},
        ]

        incar = Incar.from_file("INCAR.orig")
        incar["ICHARG"] = 10
        incar.write_file("INCAR")
        dct = handler.correct()
        assert dct["errors"] == ["aliasing_incar"]
        assert dct["actions"] == [{"action": {"_unset": {"NGY": 1, "NGZ": 1}}, "dict": "INCAR"}]


class UnconvergedErrorHandlerTest(MatSciTest):
    def setUp(self) -> None:
        copy_tmp_files(self.tmp_path, *glob("unconverged/*", root_dir=TEST_FILES))

    def test_check_correct_electronic(self) -> None:
        shutil.copy("vasprun.xml.electronic", "vasprun.xml")
        handler = UnconvergedErrorHandler()
        assert handler.check()
        dct = handler.correct()
        assert dct["errors"] == ["Unconverged"]
        assert dct == {
            "actions": [{"action": {"_set": {"ALGO": "Normal"}}, "dict": "INCAR"}],
            "errors": ["Unconverged"],
        }
        tracked_lru_cache.tracked_cache_clear()

        shutil.copy("vasprun.xml.electronic_veryfast", "vasprun.xml")
        handler = UnconvergedErrorHandler()
        assert handler.check()
        dct = handler.correct()
        assert dct["errors"] == ["Unconverged"]
        assert dct == {"actions": [{"action": {"_set": {"ALGO": "Fast"}}, "dict": "INCAR"}], "errors": ["Unconverged"]}
        tracked_lru_cache.tracked_cache_clear()

        shutil.copy("vasprun.xml.electronic_normal", "vasprun.xml")
        handler = UnconvergedErrorHandler()
        assert handler.check()
        dct = handler.correct()
        assert dct["errors"] == ["Unconverged"]
        assert dct == {
            "actions": [{"action": {"_set": {"ALGO": "All", "ISEARCH": 1}}, "dict": "INCAR"}],
            "errors": ["Unconverged"],
        }
        tracked_lru_cache.tracked_cache_clear()

        shutil.copy("vasprun.xml.electronic_metagga_fast", "vasprun.xml")
        handler = UnconvergedErrorHandler()
        assert handler.check()
        dct = handler.correct()
        assert dct["errors"] == ["Unconverged"]
        assert dct == {
            "actions": [{"action": {"_set": {"ALGO": "All", "ISEARCH": 1}}, "dict": "INCAR"}],
            "errors": ["Unconverged"],
        }
        tracked_lru_cache.tracked_cache_clear()

        shutil.copy("vasprun.xml.electronic_hybrid_fast", "vasprun.xml")
        handler = UnconvergedErrorHandler()
        assert handler.check()
        dct = handler.correct()
        assert dct["errors"] == ["Unconverged"]
        assert dct == {
            "actions": [{"action": {"_set": {"ALGO": "All", "ISEARCH": 1}}, "dict": "INCAR"}],
            "errors": ["Unconverged"],
        }
        tracked_lru_cache.tracked_cache_clear()

        shutil.copy("vasprun.xml.electronic_hybrid_all", "vasprun.xml")
        handler = UnconvergedErrorHandler()
        assert handler.check()
        dct = handler.correct()
        assert dct["errors"] == ["Unconverged"]
        assert dct["actions"] == [{"dict": "INCAR", "action": {"_set": {"ALGO": "Damped", "TIME": 0.5}}}]

    def test_check_correct_electronic_repeat(self) -> None:
        shutil.copy("vasprun.xml.electronic2", "vasprun.xml")
        handler = UnconvergedErrorHandler()
        assert handler.check()
        dct = handler.correct()
        assert dct == {
            "actions": [{"action": {"_set": {"ALGO": "All", "ISEARCH": 1}}, "dict": "INCAR"}],
            "errors": ["Unconverged"],
        }

    def test_check_correct_ionic(self) -> None:
        shutil.copy("vasprun.xml.ionic", "vasprun.xml")
        handler = UnconvergedErrorHandler()
        assert handler.check()
        dct = handler.correct()
        assert dct["errors"] == ["Unconverged"]

    def test_check_correct_scan(self) -> None:
        shutil.copy("vasprun.xml.scan", "vasprun.xml")
        handler = UnconvergedErrorHandler()
        assert handler.check()
        dct = handler.correct()
        assert dct["errors"] == ["Unconverged"]
        assert {"dict": "INCAR", "action": {"_set": {"ALGO": "All", "ISEARCH": 1}}} in dct["actions"]

    def test_amin(self) -> None:
        shutil.copy("vasprun.xml.electronic_amin", "vasprun.xml")
        handler = UnconvergedErrorHandler()
        assert handler.check()
        dct = handler.correct()
        assert dct["actions"] == [{"dict": "INCAR", "action": {"_set": {"AMIN": 0.01}}}]

    def test_as_from_dict(self) -> None:
        handler = UnconvergedErrorHandler("random_name.xml")
        h2 = UnconvergedErrorHandler.from_dict(handler.as_dict())
        assert isinstance(h2, UnconvergedErrorHandler)
        assert h2.output_filename == "random_name.xml"

    def test_correct_normal_with_condition(self) -> None:
        shutil.copy("vasprun.xml.electronic_normal", "vasprun.xml")  # Reuse an existing file
        handler = UnconvergedErrorHandler()
        assert handler.check()
        dct = handler.correct()
        assert dct["errors"] == ["Unconverged"]
        assert dct == {
            "actions": [{"action": {"_set": {"ALGO": "All", "ISEARCH": 1}}, "dict": "INCAR"}],
            "errors": ["Unconverged"],
        }

    def test_psmaxn(self) -> None:
        shutil.copy("vasprun.xml.electronic", "vasprun.xml")
        shutil.copy(f"{TEST_FILES}/large_cell_real_optlay/OUTCAR", "OUTCAR")
        handler = UnconvergedErrorHandler()
        assert handler.check()
        dct = handler.correct()
        assert set(dct["errors"]) == {"Unconverged", "psmaxn"}
        assert dct["actions"] == [
            {"action": {"_set": {"ALGO": "Normal"}}, "dict": "INCAR"},
            {"dict": "INCAR", "action": {"_set": {"LREAL": False}}},
        ]
        tracked_lru_cache.tracked_cache_clear()

    def test_uncorrectable(self) -> None:
        shutil.copy("vasprun.xml.unconverged_unfixable", "vasprun.xml")
        handler = UnconvergedErrorHandler()
        assert handler.check()
        dct = handler.correct()
        assert set(dct["errors"]) == {"Unconverged"}
        assert dct["actions"] is None


class IncorrectSmearingHandlerTest(MatSciTest):
    def setUp(self) -> None:
        copy_tmp_files(self.tmp_path, "scan_metal/INCAR", "scan_metal/vasprun.xml")

    def test_check_correct_scan_metal(self) -> None:
        handler = IncorrectSmearingHandler()
        assert handler.check()
        dct = handler.correct()
        assert dct["errors"] == ["IncorrectSmearing"]
        incar = Incar.from_file("INCAR")
        assert incar["ISMEAR"] == 2
        assert incar["SIGMA"] == 0.2


class IncorrectSmearingHandlerStaticTest(MatSciTest):
    def setUp(self) -> None:
        copy_tmp_files(self.tmp_path, "static_smearing/INCAR", "static_smearing/vasprun.xml")

    def test_check_correct_scan_metal(self) -> None:
        handler = IncorrectSmearingHandler()
        assert not handler.check()


class IncorrectSmearingHandlerFermiTest(MatSciTest):
    def setUp(self) -> None:
        copy_tmp_files(self.tmp_path, "fermi_smearing/INCAR", "fermi_smearing/vasprun.xml")

    def test_check_correct_scan_metal(self) -> None:
        handler = IncorrectSmearingHandler()
        assert not handler.check()


class KspacingMetalHandlerTest(MatSciTest):
    def setUp(self) -> None:
        copy_tmp_files(self.tmp_path, "scan_metal/INCAR", "scan_metal/vasprun.xml")

    def test_check_correct_scan_metal(self) -> None:
        handler = KspacingMetalHandler()
        assert handler.check()
        dct = handler.correct()
        assert dct["errors"] == ["ScanMetal"]
        assert Incar.from_file("INCAR")["KSPACING"] == 0.22
        os.remove("vasprun.xml")

    def test_check_with_non_kspacing_wf(self) -> None:
        os.chdir(TEST_FILES)
        shutil.copy("INCAR", f"{self.tmp_path}/INCAR")
        shutil.copy("vasprun.xml", f"{self.tmp_path}/vasprun.xml")
        handler = KspacingMetalHandler(output_filename=f"{self.tmp_path}/vasprun.xml")
        assert handler.check() is False
        os.chdir(f"{TEST_FILES}/scan_metal")

        # TODO (@janosh 2023-11-03) remove when ending ScanMetalHandler deprecation period
        assert isinstance(ScanMetalHandler(), KspacingMetalHandler)


class LargeSigmaHandlerTest(MatSciTest):
    def setUp(self) -> None:
        copy_tmp_files(self.tmp_path, *glob("large_sigma/*", root_dir=TEST_FILES))

    def test_check_correct_large_sigma(self) -> None:
        # first check should reduce sigma
        handler = LargeSigmaHandler(output_filename=zpath("OUTCAR_fail_sigma_check"))
        assert handler.check()
        dct = handler.correct()
        assert dct["errors"] == ["LargeSigma"]
        assert Incar.from_file("INCAR")["SIGMA"] == pytest.approx(0.1115, rel=1.0e-3)

        # second check should find that sigma is correct as-is
        handler = LargeSigmaHandler(output_filename=zpath("OUTCAR_pass_sigma_check"))
        assert not handler.check()

    def test_no_crash_on_partial_output(self) -> None:
        # ensure that the handler doesn't crash when the OUTCAR isn't completely written
        # this prevents jobs from being killed when the handler itself crashes

        orig_outcar_path = Path(zpath("OUTCAR_pass_sigma_check"))
        new_outcar_name = str(orig_outcar_path.parent.resolve() / f"temp_{orig_outcar_path.name}")
        shutil.copy(orig_outcar_path, new_outcar_name)

        # simulate this behavior by manually removing one of the electronic
        # entropy lines that the handler searches for
        with zopen(new_outcar_name, "rt") as f:
            data = f.read().splitlines()

        for rm_idx in range(len(data) - 1, 0, -1):
            if "T*S" in data[rm_idx]:
                data.pop(rm_idx)
                break

        with zopen(new_outcar_name, "wt") as f:
            f.write("\n".join(data))

        handler = LargeSigmaHandler(output_filename=zpath("OUTCAR_partial_output"))
        assert not handler.check()


class ZpotrfErrorHandlerTest(MatSciTest):
    def setUp(self) -> None:
        copy_tmp_files(
            self.tmp_path,
            "zpotrf/INCAR",
            "zpotrf/POSCAR",
            "zpotrf/OSZICAR.empty",
            "zpotrf/vasp.out",
            "zpotrf/OSZICAR.one_step",
        )

    def test_first_step(self) -> None:
        shutil.copy("OSZICAR.empty", "OSZICAR")
        s1 = Structure.from_file("POSCAR")
        handler = VaspErrorHandler("vasp.out")
        assert handler.check() is True
        dct = handler.correct()
        assert dct["errors"] == ["zpotrf"]
        s2 = Structure.from_file("POSCAR")
        # NOTE (@janosh on 2023-09-10) next code line used to be:
        # assert s2.volume == pytest.approx(s1.volume * 1.2**3)
        # unclear why s2.volume changed
        assert s2.volume == pytest.approx(s1.volume)
        assert s1.volume == pytest.approx(64.346221)

    def test_potim_correction(self) -> None:
        shutil.copy("OSZICAR.one_step", "OSZICAR")
        s1 = Structure.from_file("POSCAR")
        handler = VaspErrorHandler("vasp.out")
        assert handler.check() is True
        dct = handler.correct()
        assert dct["errors"] == ["zpotrf"]
        s2 = Structure.from_file("POSCAR")
        assert s2.volume == pytest.approx(s1.volume)
        assert s1.volume == pytest.approx(64.3462)
        assert Incar.from_file("INCAR")["POTIM"] == pytest.approx(0.25)

    def test_static_run_correction(self) -> None:
        shutil.copy("OSZICAR.empty", "OSZICAR")
        s1 = Structure.from_file("POSCAR")
        incar = Incar.from_file("INCAR")

        # Test for NSW 0
        incar.update({"NSW": 0})
        incar.write_file("INCAR")
        handler = VaspErrorHandler("vasp.out")
        assert handler.check() is True
        dct = handler.correct()
        assert dct["errors"] == ["zpotrf"]
        s2 = Structure.from_file("POSCAR")
        assert s2.volume == pytest.approx(s1.volume)
        assert s2.volume == pytest.approx(64.346221)
        assert Incar.from_file("INCAR")["ISYM"] == 0


class ZpotrfErrorHandlerSmallTest(MatSciTest):
    def setUp(self) -> None:
        copy_tmp_files(
            self.tmp_path,
            "zpotrf_small/INCAR",
            "zpotrf_small/POSCAR",
            "zpotrf_small/OSZICAR.empty",
            "zpotrf_small/vasp.out",
        )

    def test_small(self) -> None:
        handler = VaspErrorHandler("vasp.out")
        shutil.copy("OSZICAR.empty", "OSZICAR")
        assert handler.check()
        dct = handler.correct()
        assert dct["errors"] == ["zpotrf"]
        assert dct["actions"] == [
            {"dict": "INCAR", "action": {"_set": {"NCORE": 1}}},
            {"dict": "INCAR", "action": {"_unset": {"NPAR": 1}}},
        ]


class WalltimeHandlerTest(MatSciTest):
    def setUp(self) -> None:
        os.chdir(f"{TEST_FILES}/postprocess")
        os.environ.pop("CUSTODIAN_WALLTIME_START", None)

    def test_walltime_start(self) -> None:
        # checks the walltime handlers starttime initialization
        handler = WalltimeHandler(wall_time=3600)
        new_starttime = handler.start_time
        assert os.environ.get("CUSTODIAN_WALLTIME_START") == new_starttime.strftime("%a %b %d %H:%M:%S UTC %Y")
        # Test that walltime persists if new handler is created
        handler = WalltimeHandler(wall_time=3600)
        assert os.environ.get("CUSTODIAN_WALLTIME_START") == new_starttime.strftime("%a %b %d %H:%M:%S UTC %Y")

    def test_check_and_correct(self) -> None:
        # Try a 1 hr wall time with a 2 min buffer
        handler = WalltimeHandler(wall_time=3600, buffer_time=120)
        assert not handler.check()

        # This makes sure the check returns True when the time left is less
        # than the buffer time.
        handler.start_time = datetime.datetime.now() - datetime.timedelta(minutes=59)
        assert handler.check()

        # This makes sure the check returns True when the time left is less
        # than 3 x the average time per ionic step. We have a 62 min wall
        # time, a very short buffer time, but the start time was 62 mins ago
        handler = WalltimeHandler(wall_time=3720, buffer_time=10)
        handler.start_time = datetime.datetime.now() - datetime.timedelta(minutes=62)
        assert handler.check()

        # Test that the STOPCAR is written correctly.
        handler.correct()
        with open("STOPCAR") as file:
            content = file.read()
            assert content == "LSTOP = .TRUE."
        os.remove("STOPCAR")

        handler = WalltimeHandler(wall_time=3600, buffer_time=120, electronic_step_stop=True)

        assert not handler.check()
        handler.start_time = datetime.datetime.now() - datetime.timedelta(minutes=59)
        assert handler.check()

        handler.correct()
        with open("STOPCAR") as file:
            content = file.read()
            assert content == "LABORT = .TRUE."
        os.remove("STOPCAR")

    @classmethod
    def tearDown(cls) -> None:
        os.environ.pop("CUSTODIAN_WALLTIME_START", None)
        os.chdir(CWD)


class PositiveEnergyHandlerTest(MatSciTest):
    def setUp(self) -> None:
        copy_tmp_files(self.tmp_path, "positive_energy/INCAR", "positive_energy/POSCAR", "positive_energy/OSZICAR")

    def test_check_correct(self) -> None:
        handler = PositiveEnergyErrorHandler()
        assert handler.check()
        dct = handler.correct()
        assert dct["errors"] == ["Positive energy"]

        assert os.path.isfile("error.1.tar.gz")
        with tarfile.open("error.1.tar.gz", "r:gz") as tar:
            assert len(tar.getmembers()) > 0

        incar = Incar.from_file("INCAR")

        assert incar["ALGO"] == "Normal"


class PotimHandlerTest(MatSciTest):
    def setUp(self) -> None:
        copy_tmp_files(self.tmp_path, "potim/INCAR", "potim/POSCAR", "potim/OSZICAR")

    def test_check_correct(self) -> None:
        incar = Incar.from_file("INCAR")
        original_potim = incar["POTIM"]

        handler = PotimErrorHandler()
        assert handler.check()
        dct = handler.correct()
        assert dct["errors"] == ["POTIM"]

        assert os.path.isfile("error.1.tar.gz")

        incar = Incar.from_file("INCAR")
        new_potim = incar["POTIM"]

        assert original_potim == new_potim
        assert incar["IBRION"] == 3


class LrfCommHandlerTest(MatSciTest):
    def setUp(self) -> None:
        copy_tmp_files(self.tmp_path, "lrf_comm/INCAR", "lrf_comm/OUTCAR", "lrf_comm/std_err.txt")

    def test_lrf_comm(self) -> None:
        handler = LrfCommutatorHandler("std_err.txt")
        assert handler.check() is True
        dct = handler.correct()
        assert dct["errors"] == ["lrf_comm"]
        vi = VaspInput.from_directory(".")
        assert vi["INCAR"]["LPEAD"] is True


class KpointsTransHandlerTest(MatSciTest):
    def setUp(self) -> None:
        copy_tmp_files(self.tmp_path, "KPOINTS", "std_err.txt.kpoints_trans")

    def test_kpoints_trans(self) -> None:
        handler = StdErrHandler("std_err.txt.kpoints_trans")
        assert handler.check() is True
        dct = handler.correct()
        assert dct["errors"] == ["kpoints_trans"]
        assert dct["actions"] == [{"action": {"_set": {"kpoints": [[4, 4, 4]]}}, "dict": "KPOINTS"}]

        assert handler.check() is True
        dct = handler.correct()
        assert dct["errors"] == ["kpoints_trans"]
        assert dct["actions"] == []  # don't correct twice


class OutOfMemoryHandlerTest(MatSciTest):
    def setUp(self) -> None:
        copy_tmp_files(self.tmp_path, "INCAR", "std_err.txt.oom")

    def test_oom(self) -> None:
        vi = VaspInput.from_directory(".")

        VaspModder(vi=vi).apply_actions([{"dict": "INCAR", "action": {"_set": {"KPAR": 4}}}])
        handler = StdErrHandler("std_err.txt.oom")
        assert handler.check() is True
        dct = handler.correct()
        assert dct["errors"] == ["out_of_memory"]
        assert dct["actions"] == [{"dict": "INCAR", "action": {"_set": {"KPAR": 2}}}]


class DriftErrorHandlerTest(MatSciTest):
    def setUp(self) -> None:
        copy_tmp_files(self.tmp_path, "INCAR", "drift/OUTCAR", "drift/CONTCAR")

    def test_check(self) -> None:
        handler = DriftErrorHandler(max_drift=0.05, to_average=11)
        assert not handler.check()

        handler = DriftErrorHandler(max_drift=0.05)
        assert not handler.check()

        handler = DriftErrorHandler(max_drift=0.0001)
        assert not handler.check()

        incar = Incar.from_file("INCAR")
        incar["EDIFFG"] = -0.01
        incar.write_file("INCAR")

        handler = DriftErrorHandler(max_drift=0.0001)
        assert handler.check()

        handler = DriftErrorHandler()
        handler.check()
        assert handler.max_drift == 0.01

    def test_correct(self) -> None:
        handler = DriftErrorHandler(max_drift=0.0001, enaug_multiply=2)
        handler.check()
        handler.correct()
        incar = Incar.from_file("INCAR")
        assert incar.get("PREC") == "High"
        assert incar.get("ENAUG", 0) == incar.get("ENCUT", 2) * 2


class NonConvergingErrorHandlerTest(MatSciTest):
    n_ionic_steps: int = 3

    def setUp(self) -> None:
        copy_tmp_files(self.tmp_path, *glob("nonconv/*", root_dir=TEST_FILES))

    def test_check(self) -> None:
        # calculation has four ionic steps which each hit NELM = 10
        handler = NonConvergingErrorHandler(nionic_steps=self.n_ionic_steps)
        assert handler.check()

        # increase NELM to avoid NonConvergingErrorHandler
        incar = Incar.from_file("INCAR")
        incar["NELM"] = 15
        incar.write_file("INCAR")
        assert not handler.check()

    def test_correct(self) -> None:
        original_incar = Incar.from_file("INCAR")

        handler = NonConvergingErrorHandler(nionic_steps=self.n_ionic_steps)
        handler.check()

        # INCAR has ALGO = Fast, so first correction --> Normal
        handler.correct()
        incar = Incar.from_file("INCAR")
        assert incar["ALGO"].lower() == "normal"

        # because ISMEAR = -5, skip ALGO = all and adjust
        post_all_corrections = {"ALGO": "Normal", "AMIX": 0.1, "BMIX": 0.01, "ICHARG": 2}
        handler.correct()
        incar = Incar.from_file("INCAR")
        assert all(value == incar[key] for key, value in post_all_corrections.items())

        incar.update({"AMIX": 0.02, "BMIX": 2.9})
        post_all_corrections = {"ALGO": "Normal", "AMIN": 0.01, "BMIX": 3.0, "ICHARG": 2}
        handler.correct()
        incar = Incar.from_file("INCAR")
        assert all(value == incar[key] for key, value in post_all_corrections.items())

        # now replace ISMEAR --> 0, ALGO --> VeryFast to get ladder
        incar = Incar(original_incar)  # incar.copy() returns dict
        incar.update({"ISMEAR": 0, "ALGO": "veryfast"})
        incar.write_file("INCAR")

        algo_ladder = ("fast", "normal", "all")
        for algo in algo_ladder:
            handler.correct()
            incar = Incar.from_file("INCAR")
            assert incar["ALGO"].lower() == algo

        # now test meta-GGA and hybrid, should go directly from ALGO = fast to all
        for updates in ({"METAGGA": "SCAN"}, {"LHFCALC": True, "GGA": "PE"}):
            incar = Incar(original_incar)  # incar.copy() returns dict
            incar.update(updates)
            incar.write_file("INCAR")
            handler.correct()

            incar = Incar.from_file("INCAR")
            assert incar["ALGO"].lower() == "all"

    def test_as_from_dict(self) -> None:
        handler = NonConvergingErrorHandler("OSZICAR_random")
        h2 = NonConvergingErrorHandler.from_dict(handler.as_dict())
        assert isinstance(h2, NonConvergingErrorHandler)
        assert h2.output_filename == "OSZICAR_random"



"""This module contains utility functions that are useful for Q-Chem jobs."""

import math

import numpy as np


def perturb_coordinates(old_coords, negative_freq_vecs, molecule_perturb_scale, reversed_direction):
    """
    Perturbs a structure along the imaginary mode vibrational frequency vectors.

    old_coords (np.ndarray): Initial molecule coordinates
    negative_freq_vecs (list of np.ndarray): Vibrational frequency vectors corresponding to
        imaginary (negative) frequencies
    molecule_perturb_scale (float): Scaling factor for coordination modification. The perturbation
        vector will be multiplied by this factor.
    reversed_direction (bool): If True, then perturb in direction opposite of frequency modes.

    """
    max_dis = max(math.sqrt(sum(x**2 for x in vec)) for vec in negative_freq_vecs)
    scale = molecule_perturb_scale / max_dis
    normalized_vecs = [[x * scale for x in vec] for vec in negative_freq_vecs]
    direction = 1.0
    if reversed_direction:
        direction = -1.0
    return [
        [c + v * direction for c, v in zip(coord, vec, strict=False)]
        for coord, vec in zip(old_coords, normalized_vecs, strict=False)
    ]


def vector_list_diff(vecs1, vecs2):
    """
    Calculates the summed difference of two vectors.

    Typically this function is used to compare between the different atom-wise
    components of a vibrational frequency mode vector.

    vecs1 (np.ndarray): Collection of vectors to be compared. Typical dimension
        n x 3
    vecs2 (np.ndarray): Collection of vectors to be compared. Typical dimension
        n x 3
    """
    diff = 0.0
    if len(vecs1) != len(vecs2):
        raise AssertionError("ERROR: Vectors must be of equal length! Exiting...")
    for ii, vec1 in enumerate(vecs1):
        diff += np.linalg.norm(vecs2[ii] - vec1)
    return diff



import glob
import gzip
import os
import shutil
from unittest import TestCase

from custodian.gaussian.jobs import GaussianJob
from tests.conftest import TEST_FILES

__author__ = "Rasha Atwi"
__version__ = "0.1"
__maintainer__ = "Rasha Atwi"
__email__ = "rasha.atwi@stonybrook.edu"
__status__ = "Alpha"
__date__ = "3/21/24"

TEST_DIR = f"{TEST_FILES}/gaussian"
SCR_DIR = f"{TEST_DIR}/scratch"
CWD = os.getcwd()


class TestGaussianJob(TestCase):
    def setUp(self):
        self.input_file = "test.com"
        self.output_file = "test.out"
        self.gaussian_cmd = f"g16 < {self.input_file} > {self.output_file}"
        self.stderr_file = "stderr.txt"
        self.suffix = ".test"
        self.backup = True
        self.directory = SCR_DIR

        os.makedirs(SCR_DIR, exist_ok=True)
        shutil.copyfile(f"{TEST_DIR}/mol_opt.com", f"{SCR_DIR}/test.com")
        os.chdir(SCR_DIR)

    def tearDown(self):
        os.chdir(CWD)
        shutil.rmtree(SCR_DIR)
        files_to_remove = glob.glob(f"{TEST_DIR}/*.out")
        if files_to_remove and glob.glob(f"{TEST_DIR}/*.out.gz"):
            for file_path in files_to_remove:
                os.remove(file_path)

    def test_normal(self):
        job = GaussianJob(
            self.gaussian_cmd,
            self.input_file,
            self.output_file,
            self.stderr_file,
            self.suffix,
            self.backup,
        )
        job.setup()
        assert os.path.exists(f"{SCR_DIR}/test.com.orig")
        if not os.path.exists(f"{TEST_DIR}/mol_opt.out") and os.path.exists(f"{TEST_DIR}/mol_opt.out.gz"):
            with gzip.open(f"{TEST_DIR}/mol_opt.out.gz", "rb") as f_in, open(f"{TEST_DIR}/mol_opt.out", "wb") as f_out:
                shutil.copyfileobj(f_in, f_out)
        shutil.copy(f"{TEST_DIR}/mol_opt.out", f"{SCR_DIR}/test.out")
        job.postprocess()
        assert os.path.exists(f"{SCR_DIR}/test.com{self.suffix}")
        assert os.path.exists(f"{SCR_DIR}/test.out{self.suffix}")

    def test_better_guess(self):
        job_gen = GaussianJob.generate_better_guess(
            self.gaussian_cmd,
            self.input_file,
            self.output_file,
            self.stderr_file,
            self.backup,
            cart_coords=True,
            directory=self.directory,
        )
        jobs = list(job_gen)
        assert len(jobs) == 1, "One job should be generated under normal conditions."
        jobs[0].setup()
        assert os.path.exists(f"{SCR_DIR}/test.com.orig")
        if not os.path.exists(f"{TEST_DIR}/mol_opt.out") and os.path.exists(f"{TEST_DIR}/mol_opt.out.gz"):
            with gzip.open(f"{TEST_DIR}/mol_opt.out.gz", "rb") as f_in, open(f"{TEST_DIR}/mol_opt.out", "wb") as f_out:
                shutil.copyfileobj(f_in, f_out)
        shutil.copy(f"{TEST_DIR}/mol_opt.out", f"{SCR_DIR}/test.out")
        jobs[0].postprocess()
        assert os.path.exists(f"{SCR_DIR}/test.com.guess1")
        assert os.path.exists(f"{SCR_DIR}/test.out.guess1")



import os
import shutil
import unittest
from unittest import TestCase

from pymatgen.io.qchem.inputs import QCInput

from custodian.qchem.handlers import QChemErrorHandler
from custodian.qchem.jobs import QCJob
from tests.conftest import TEST_FILES

try:
    from openbabel import openbabel as ob
except ImportError:
    ob = None

__author__ = "Samuel Blau"
__copyright__ = "Copyright 2022, The Materials Project"
__version__ = "0.1"
__maintainer__ = "Samuel Blau"
__email__ = "samblau1@gmail.com"
__status__ = "Alpha"
__date__ = "6/3/22"

TEST_DIR = f"{TEST_FILES}/qchem/new_test_files"
SCR_DIR = f"{TEST_DIR}/scratch"
CWD = os.getcwd()
skip_if_no_openbabel = unittest.skipIf(ob is None, "openbabel not installed")


@skip_if_no_openbabel
class FFOptJobHandlerInteraction(TestCase):
    def _check_equivalent_inputs(self, input1, input2) -> None:
        QCinput1 = QCInput.from_file(input1)
        QCinput2 = QCInput.from_file(input2)
        sections1 = QCInput.find_sections(QCinput1.get_str())
        sections2 = QCInput.find_sections(QCinput2.get_str())
        assert sections1 == sections2
        for key in sections1:
            assert QCinput1.as_dict().get(key) == QCinput2.as_dict().get(key)

    def setUp(self) -> None:
        os.makedirs(f"{SCR_DIR}/scratch", exist_ok=True)
        shutil.copyfile(f"{TEST_DIR}/job_handler_interaction/mol.qin.orig", f"{SCR_DIR}/mol.qin")
        shutil.copyfile(f"{TEST_DIR}/job_handler_interaction/error.1/mol.qout", f"{SCR_DIR}/mol.qout.error1")
        shutil.copyfile(f"{TEST_DIR}/job_handler_interaction/error.2/mol.qin", f"{SCR_DIR}/mol.qin.error2")
        shutil.copyfile(f"{TEST_DIR}/job_handler_interaction/error.2/mol.qout", f"{SCR_DIR}/mol.qout.error2")
        shutil.copyfile(f"{TEST_DIR}/job_handler_interaction/error.3/mol.qin", f"{SCR_DIR}/mol.qin.error3")
        shutil.copyfile(f"{TEST_DIR}/job_handler_interaction/error.3/mol.qout", f"{SCR_DIR}/mol.qout.error3")
        shutil.copyfile(f"{TEST_DIR}/job_handler_interaction/mol.qin.opt_0", f"{SCR_DIR}/mol.qin.opt_0")
        shutil.copyfile(f"{TEST_DIR}/job_handler_interaction/mol.qout.opt_0", f"{SCR_DIR}/mol.qout.opt_0")
        shutil.copyfile(f"{TEST_DIR}/job_handler_interaction/error.5/mol.qin", f"{SCR_DIR}/mol.qin.error5")
        shutil.copyfile(f"{TEST_DIR}/job_handler_interaction/error.5/mol.qout", f"{SCR_DIR}/mol.qout.error5")
        shutil.copyfile(f"{TEST_DIR}/job_handler_interaction/mol.qin.freq_0", f"{SCR_DIR}/mol.qin.freq_0")
        shutil.copyfile(f"{TEST_DIR}/job_handler_interaction/mol.qout.freq_0", f"{SCR_DIR}/mol.qout.freq_0")
        shutil.copyfile(f"{TEST_DIR}/job_handler_interaction/mol.qin.opt_1", f"{SCR_DIR}/mol.qin.opt_1")
        os.chdir(SCR_DIR)

    def tearDown(self) -> None:
        os.chdir(CWD)
        shutil.rmtree(SCR_DIR)

    def test_OptFF(self) -> None:
        job = QCJob.opt_with_frequency_flattener(
            qchem_command="qchem",
            max_cores=40,
            input_file="mol.qin",
            output_file="mol.qout",
            linked=True,
        )
        expected_next = QCJob(
            qchem_command="qchem",
            max_cores=40,
            multimode="openmp",
            input_file="mol.qin",
            output_file="mol.qout",
            suffix=".opt_0",
            save_scratch=True,
            backup=True,
        ).as_dict()
        assert next(job).as_dict() == expected_next

        handler = QChemErrorHandler(
            input_file="mol.qin",
            output_file="mol.qout.error1",
        )
        handler.check()
        dct = handler.correct()
        assert dct["errors"] == ["back_transform_error"]
        assert dct["actions"] == [{"molecule": "molecule_from_last_geometry"}]
        self._check_equivalent_inputs("mol.qin", "mol.qin.error2")

        handler = QChemErrorHandler(
            input_file="mol.qin",
            output_file="mol.qout.error2",
        )
        handler.check()
        dct = handler.correct()
        assert dct["errors"] == ["SCF_failed_to_converge"]
        assert dct["actions"] == [{"scf_algorithm": "gdm"}, {"max_scf_cycles": "500"}]
        self._check_equivalent_inputs("mol.qin", "mol.qin.error3")

        handler = QChemErrorHandler(
            input_file="mol.qin",
            output_file="mol.qout.error3",
        )
        handler.check()
        dct = handler.correct()
        assert dct["errors"] == ["back_transform_error"]
        assert dct["actions"] == [{"molecule": "molecule_from_last_geometry"}]
        self._check_equivalent_inputs("mol.qin", "mol.qin.opt_0")

        handler = QChemErrorHandler(
            input_file="mol.qin",
            output_file="mol.qout.opt_0",
        )
        handler.check()
        assert not handler.check()

        expected_next = QCJob(
            qchem_command="qchem",
            max_cores=40,
            multimode="openmp",
            input_file="mol.qin",
            output_file="mol.qout",
            suffix=".freq_0",
            save_scratch=True,
            backup=False,
        ).as_dict()
        assert next(job).as_dict() == expected_next
        self._check_equivalent_inputs("mol.qin", "mol.qin.error5")

        handler = QChemErrorHandler(
            input_file="mol.qin",
            output_file="mol.qout.error5",
        )
        handler.check()
        dct = handler.correct()
        assert dct["errors"] == ["failed_cpscf"]
        assert dct["actions"] == [{"cpscf_nseg": "3"}]
        self._check_equivalent_inputs("mol.qin", "mol.qin.freq_0")

        handler = QChemErrorHandler(
            input_file="mol.qin",
            output_file="mol.qout.freq_0",
        )
        handler.check()
        assert not handler.check()

        expected_next = QCJob(
            qchem_command="qchem",
            max_cores=40,
            multimode="openmp",
            input_file="mol.qin",
            output_file="mol.qout",
            suffix=".opt_1",
            save_scratch=True,
            backup=False,
        ).as_dict()
        assert next(job).as_dict() == expected_next

        self._check_equivalent_inputs("mol.qin", "mol.qin.opt_1")



"""Implements various interpreters and modders for FEFF calculations."""

import os

from pymatgen.io.feff.sets import FEFFDictSet

from custodian.ansible.actions import DictActions, FileActions
from custodian.ansible.interpreter import Modder


class FeffModder(Modder):
    """A Modder for FeffInput sets."""

    def __init__(self, actions=None, strict=True, feffinp=None, directory="./") -> None:
        """Initialize a FeffModder.

        Args:
            actions ([Action]): A sequence of supported actions. See
            actions ([Action]): A sequence of supported actions. See
                :mod:`custodian.ansible.actions`. Default is None,
                which means DictActions and FileActions are supported.
            strict (bool): Indicating whether to use strict mode. In non-strict
                mode, unsupported actions are simply ignored without any
                errors raised. In strict mode, if an unsupported action is
                supplied, a ValueError is raised. Defaults to True.
            feffinp (FEFFInput): A FeffInput object from the current directory.
                Initialized automatically if not passed (but passing it will
                avoid having to re-parse the directory).
            directory (str): The directory containing the FeffInput set. Defaults to "./".
        """
        self.directory = directory
        self.feffinp = feffinp or FEFFDictSet.from_directory(self.directory)
        self.feffinp = self.feffinp.all_input()
        actions = actions or [FileActions, DictActions]
        super().__init__(actions, strict, directory=directory)

    def apply_actions(self, actions) -> None:
        """
        Applies a list of actions to the FEFF Input Set and rewrites modified
        files.

        Args:
            actions (dict): A list of actions of the form {'file': filename,
                'action': moddermodification} or {'dict': feffinput_key,
                'action': moddermodification}
        """
        modified = []
        for action in actions:
            if "dict" in action:
                key = action["dict"]
                modified.append(key)
                self.feffinp[key] = self.modify_object(action["action"], self.feffinp[key])  # type:ignore[index]
            elif "file" in action:
                self.modify(action["action"], action["file"])
            else:
                raise ValueError(f"Unrecognized format: {action}")
        if modified:
            feff = self.feffinp
            feff_input = "\n\n".join(
                str(feff[key])  # type:ignore[index,operator]
                for key in ("HEADER", "PARAMETERS", "POTENTIALS", "ATOMS")
                if key in feff  # type:ignore[index,operator]
            )
            for key, val in feff.items():  # type:ignore[union-attr]
                with open(os.path.join(self.directory, key), "w") as file:
                    file.write(str(val))

            with open(os.path.join(self.directory, "feff.inp"), "w") as file:
                file.write(feff_input)
